from typing import Dict, Union, List

from .types import (
    Quirks,
    RobotMountConfigs,
    AvailableUnits,
    PipetteGenerationType,
    PipetteChannelType,
)

MOUNT_CONFIG_LOOKUP_TABLE = {
    PipetteGenerationType.GEN1: {
        PipetteChannelType.SINGLE_CHANNEL: RobotMountConfigs(768, 220, 30),
        PipetteChannelType.EIGHT_CHANNEL: RobotMountConfigs(768, 220, 30),
    },
    PipetteGenerationType.GEN2: {
        PipetteChannelType.SINGLE_CHANNEL: RobotMountConfigs(3200, 172.15, 60),
        PipetteChannelType.EIGHT_CHANNEL: RobotMountConfigs(3200, 155.75, 60),
    },
    PipetteGenerationType.FLEX: {
        PipetteChannelType.SINGLE_CHANNEL: RobotMountConfigs(2133.33, 230.15, 80),
        PipetteChannelType.EIGHT_CHANNEL: RobotMountConfigs(2133.33, 230.15, 80),
        PipetteChannelType.NINETY_SIX_CHANNEL: RobotMountConfigs(2133.33, 230.15, 80),
    },
}


VALID_QUIRKS = [q.value for q in Quirks]
MUTABLE_CONFIGS_V1 = [
    "top",
    "bottom",
    "blowout",
    "dropTip",
    "pickUpCurrent",
    "pickUpDistance",
    "pickUpIncrement",
    "pickUpPresses",
    "pickUpSpeed",
    "plungerCurrent",
    "dropTipCurrent",
    "dropTipSpeed",
    "tipLength",
    "quirks",
]

# Version 2 Configuration Keys.
MUTABLE_CONFIGS_V2 = [
    "plungerPositionsConfigurations",
    "pickUpTipConfigurations",
    "dropTipConfigurations",
    "plungerMotorConfigurations",
    "tipLength",
]

RESTRICTED_MUTABLE_CONFIG_KEYS = [*VALID_QUIRKS, "model"]


_MAP_KEY_TO_V2: Dict[str, List[str]] = {
    "top": ["plungerPositionsConfigurations", "default", "top"],
    "bottom": ["plungerPositionsConfigurations", "default", "bottom"],
    "blowout": ["plungerPositionsConfigurations", "default", "blowout"],
    "dropTip": ["plungerPositionsConfigurations", "default", "drop"],
    "pickUpCurrent": [
        "pickUpTipConfigurations",
        "pressFit",
        "configurationsByNozzleMap",
        "##EACHNOZZLEMAP##",
        "##EACHTIPTYPE##",
        "current",
    ],
    "pickUpDistance": [
        "pickUpTipConfigurations",
        "pressFit",
        "configurationsByNozzleMap",
        "##EACHNOZZLEMAP##",
        "##EACHTIPTYPE##",
        "distance",
    ],
    "pickUpIncrement": ["pickUpTipConfigurations", "pressFit", "increment"],
    "pickUpPresses": ["pickUpTipConfigurations", "pressFit", "presses"],
    "pickUpSpeed": [
        "pickUpTipConfigurations",
        "pressFit",
        "configurationsByNozzleMap",
        "##EACHNOZZLEMAP##",
        "##EACHTIPTYPE##",
        "speed",
    ],
    "plungerCurrent": ["plungerMotorConfigurations", "run"],
    "dropTipCurrent": ["dropTipConfigurations", "plungerEject", "current"],
    "dropTipSpeed": ["dropTipConfigurations", "plungerEject", "speed"],
    "maxVolume": ["liquid_properties", "default", "maxVolume"],
    "minVolume": ["liquid_properties", "default", "minVolume"],
    "tipLength": [
        "liquid_properties",
        "default",
        "supportedTips",
        "##EACHTIP##",
        "defaultTipLength",
    ],
}


_MIN_MAX_LOOKUP: Dict[str, Dict[str, Union[int, float]]] = {
    "current": {"min": 0.1, "max": 2.0},
    "run": {"min": 0.1, "max": 2.0},
    "speed": {"min": 0.01, "max": 30},
    "top": {"min": -20, "max": 30},
    "bottom": {"min": -20, "max": 30},
    "blowout": {"min": -20, "max": 30},
    "drop": {"min": -20, "max": 30},
    "presses": {"min": 0, "max": 15},
    "increment": {"min": 0, "max": 10},
    "distance": {"min": 0, "max": 10},
    "defaultTipLength": {"min": 0, "max": 100},
}

_TYPE_LOOKUP: Dict[str, str] = {
    "current": "float",
    "run": "float",
    "speed": "float",
    "top": "float",
    "bottom": "float",
    "blowout": "float",
    "drop": "float",
    "presses": "int",
    "increment": "int",
    "defaultTipLength": "float",
    "distance": "float",
}

_UNITS_LOOKUP = {
    "current": AvailableUnits.amps,
    "run": AvailableUnits.amps,
    "speed": AvailableUnits.speed,
    "top": AvailableUnits.mm,
    "bottom": AvailableUnits.mm,
    "blowout": AvailableUnits.mm,
    "drop": AvailableUnits.mm,
    "presses": AvailableUnits.presses,
    "increment": AvailableUnits.mm,
    "defaultTipLength": AvailableUnits.mm,
    "distance": AvailableUnits.mm,
}



"""opentrons_shared_data.gripper: functions and types for gripper config."""
from typing import cast, Any
from typing_extensions import Literal
import json
from pathlib import Path

from .. import load_shared_data
from .gripper_definition import (
    GripperDefinition,
    GripperSchema,
    GripperSchemaVersion,
    GripForceProfile,
    GripperModel,
    Geometry,
)


class InvalidGripperDefinition(Exception):
    """Incorrect gripper definition."""

    pass


def generate_schema() -> str:
    """Create schema."""
    raw_json_schema = GripperDefinition.schema_json()
    schema_as_dict = json.loads(raw_json_schema)
    return json.dumps(schema_as_dict, indent=2)


def load_schema(version: Literal[1]) -> GripperSchema:
    """Load gripper schema."""
    path = Path("gripper") / "schemas" / f"{version}.json"
    return cast(GripperSchema, json.loads(load_shared_data(path)))


def _load_definition_data(path: Path) -> Any:
    return


def load_definition(
    model: GripperModel,
    version: GripperSchemaVersion = 1,
) -> GripperDefinition:
    """Load gripper definition based on schema version and gripper model."""
    try:
        path = Path("gripper") / "definitions" / f"{version}" / f"{model.value}.json"
        return GripperDefinition.model_validate(json.loads(load_shared_data(path)))
    except FileNotFoundError:
        raise InvalidGripperDefinition(
            f"Gripper model {model} definition in schema version {version} does not exist."
        )


__all__ = [
    "GripperDefinition",
    "GripperSchema",
    "GripperSchemaVersion",
    "GripForceProfile",
    "GripperModel",
    "Geometry",
    "generate_schema",
]



import pytest
from typing import cast, List
from opentrons_shared_data.pipette.types import (
    PipetteChannelType,
    PipetteModelType,
    PipetteVersionType,
    PipetteModelMajorVersionType,
    PipetteModelMinorVersionType,
    LiquidClasses,
)
from opentrons_shared_data.pipette.pipette_definition import (
    liquid_class_for_volume_between_default_and_defaultlowvolume,
    PipetteLiquidPropertiesDefinition,
    SupportedTipsDefinition,
)


def get_liquid_definition_for(
    liquid_class: LiquidClasses,
) -> PipetteLiquidPropertiesDefinition:
    if liquid_class == LiquidClasses.lowVolumeDefault:
        return PipetteLiquidPropertiesDefinition.model_validate(
            {
                "supportedTips": {
                    "t50": SupportedTipsDefinition.model_validate(
                        {
                            "defaultAspirateFlowRate": {
                                "default": 5,
                                "valuesByApiLevel": {"2.0": 5},
                            },
                            "defaultDispenseFlowRate": {
                                "default": 5,
                                "valuesByApiLevel": {"2.0": 5},
                            },
                            "defaultBlowOutFlowRate": {
                                "default": 10,
                                "valuesByApiLevel": {"2.0": 10},
                            },
                            "defaultFlowAcceleration": 100,
                            "defaultTipLength": 100,
                            "defaultReturnTipHeight": 0.5,
                            "defaultPushOutVolume": 5,
                            "aspirate": {"default": {"1": [[1, 2, 3]]}},
                            "dispense": {"default": {"1": [[1, 2, 3]]}},
                        }
                    )
                },
                "versionedTipOverlapDictionary": {"v0": {}},
                "maxVolume": 30,
                "minVolume": 1,
                "defaultTipracks": [],
            }
        )
    else:
        return PipetteLiquidPropertiesDefinition.model_validate(
            {
                "supportedTips": {
                    "t50": SupportedTipsDefinition.model_validate(
                        {
                            "defaultAspirateFlowRate": {
                                "default": 5,
                                "valuesByApiLevel": {"2.0": 5},
                            },
                            "defaultDispenseFlowRate": {
                                "default": 5,
                                "valuesByApiLevel": {"2.0": 5},
                            },
                            "defaultBlowOutFlowRate": {
                                "default": 10,
                                "valuesByApiLevel": {"2.0": 10},
                            },
                            "defaultFlowAcceleration": 100,
                            "defaultTipLength": 100,
                            "defaultReturnTipHeight": 0.5,
                            "defaultPushOutVolume": 5,
                            "aspirate": {"default": {"1": [[1, 2, 3]]}},
                            "dispense": {"default": {"1": [[1, 2, 3]]}},
                        }
                    )
                },
                "versionedTipOverlapDictionary": {"v0": {}},
                "maxVolume": 100,
                "minVolume": 5,
                "defaultTipracks": [],
            }
        )


@pytest.mark.parametrize(
    argnames=["model", "expected_enum"],
    argvalues=[["p50", PipetteModelType.p50], ["p1000", PipetteModelType.p1000]],
)
def test_model_enum(model: str, expected_enum: PipetteModelType) -> None:
    assert expected_enum == PipetteModelType(model)


@pytest.mark.parametrize(argnames="channels", argvalues=[1, 8, 96])
def test_channel_enum(channels: int) -> None:
    channel_type = PipetteChannelType(channels)
    assert channels == channel_type


def test_incorrect_values() -> None:
    with pytest.raises(ValueError):
        PipetteModelType("p100")

    with pytest.raises(ValueError):
        PipetteChannelType(99)


@pytest.mark.parametrize(
    argnames=["major", "minor"],
    argvalues=[[1, 0], [1, 3], [3, 9]],
)
def test_version_enum(major: int, minor: int) -> None:
    version_type = PipetteVersionType(
        cast(PipetteModelMajorVersionType, major),
        cast(PipetteModelMinorVersionType, minor),
    )
    assert version_type.as_tuple == (major, minor)


@pytest.mark.parametrize(
    argnames=[
        "all_liquid_classes",
        "volume",
        "current_liquid_class",
        "expected_liquid_class",
    ],
    argvalues=[
        [
            [LiquidClasses.default],
            1,
            LiquidClasses.lowVolumeDefault,
            LiquidClasses.default,
        ],
        [
            [LiquidClasses.lowVolumeDefault, LiquidClasses.default],
            1,
            LiquidClasses.lowVolumeDefault,
            LiquidClasses.lowVolumeDefault,
        ],
        [
            [LiquidClasses.lowVolumeDefault, LiquidClasses.default],
            1,
            LiquidClasses.default,
            LiquidClasses.lowVolumeDefault,
        ],
        [
            [LiquidClasses.lowVolumeDefault, LiquidClasses.default],
            5,
            LiquidClasses.default,
            LiquidClasses.default,
        ],
        [
            [LiquidClasses.lowVolumeDefault, LiquidClasses.default],
            5,
            LiquidClasses.lowVolumeDefault,
            LiquidClasses.default,
        ],
        [
            [LiquidClasses.lowVolumeDefault, LiquidClasses.default],
            100,
            LiquidClasses.lowVolumeDefault,
            LiquidClasses.default,
        ],
    ],
)
def test_liquid_class_for_volume_between_default_and_defaultlowvolume(
    all_liquid_classes: List[LiquidClasses],
    volume: float,
    current_liquid_class: LiquidClasses,
    expected_liquid_class: LiquidClasses,
) -> None:
    available_liquid_classes = {
        lc: get_liquid_definition_for(lc) for lc in all_liquid_classes
    }
    assert (
        liquid_class_for_volume_between_default_and_defaultlowvolume(
            volume, current_liquid_class, available_liquid_classes
        )
        == expected_liquid_class
    )



"""
This was initially generated by datamodel-codegen from the labware schema in
shared-data. It's been modified by hand to be more friendly.
"""

from __future__ import annotations

from enum import Enum
from math import sqrt, asin
from typing import Final
from numpy import pi, trapz
from functools import cached_property

from pydantic import (
    ConfigDict,
    BaseModel,
    Discriminator,
    Field,
    StrictInt,
    StrictFloat,
    TypeAdapter,
)
from typing_extensions import Annotated, Literal

from .types import LocatingFeatures
from .constants import (
    Conical,
    Cuboidal,
    RoundedCuboid,
    SquaredCone,
    Spherical,
    WellShape,
    Circular,
    Rectangular,
)

SAFE_STRING_REGEX = "^[a-z0-9._]+$"
RECURSIVE_SEARCH_VOLUME_TOLERANCE = 0.001

# TODO(jh, 2025-05-09): We need to handle both positive numbers (by schema 2 convention)
#  and negative numbers (by schema 3 convention) no matter what, so we can just not enforce signage at the schema level.

_StrictNonNegativeInt = Annotated[int, Field(strict=True, ge=0)]
_StrictNonNegativeFloat = Annotated[float, Field(strict=True, ge=0.0)]
_StrictNonPositiveInt = Annotated[int, Field(strict=True, le=0)]
_StrictNonPositiveFloat = Annotated[float, Field(strict=True, le=0.0)]


_Number = StrictInt | StrictFloat
"""JSON number type, written to preserve lack of decimal point.

For labware definition hashing, which is an older part of the codebase,
this ensures that Pydantic won't change `"someFloatField: 0` to
`"someFloatField": 0.0`, which would hash differently.
"""

_NonNegativeNumber = _StrictNonNegativeInt | _StrictNonNegativeFloat
"""Non-negative JSON number type, written to preserve lack of decimal point."""

_NonPositiveNumber = _StrictNonPositiveInt | _StrictNonPositiveFloat
"""Non-positive JSON number type, written to preserve lack of decimal point."""


class Vector2D(BaseModel):
    x: _Number
    y: _Number


class Vector3D(BaseModel):
    x: _Number
    y: _Number
    z: _Number


class AxisAlignedBoundingBox2D(BaseModel):
    backLeft: Vector2D
    frontRight: Vector2D


class AxisAlignedBoundingBox3D(BaseModel):
    backLeftBottom: Vector3D
    frontRightTop: Vector3D


class GripperOffsets(BaseModel):
    pickUpOffset: Vector3D
    dropOffset: Vector3D


class BrandData(BaseModel):
    brand: str
    brandId: list[str] | None = None
    links: list[str] | None = None


class DisplayCategory(str, Enum):
    tipRack = "tipRack"
    tubeRack = "tubeRack"
    reservoir = "reservoir"
    trash = "trash"
    wellPlate = "wellPlate"
    aluminumBlock = "aluminumBlock"
    adapter = "adapter"
    other = "other"
    lid = "lid"
    system = "system"


class LabwareRole(str, Enum):
    labware = "labware"
    fixture = "fixture"
    adapter = "adapter"
    maintenance = "maintenance"
    lid = "lid"
    system = "system"


class Metadata(BaseModel):
    displayName: str
    displayCategory: DisplayCategory
    displayVolumeUnits: Literal["µL", "mL", "L"]
    tags: list[str] | None = None


class Parameters2(BaseModel):
    format: Literal["96Standard", "384Standard", "trough", "irregular", "trash"]
    quirks: list[str] | None = None
    isTiprack: bool
    tipLength: _NonNegativeNumber | None = None
    tipOverlap: _NonNegativeNumber | None = None
    loadName: Annotated[str, Field(pattern=SAFE_STRING_REGEX)]
    isMagneticModuleCompatible: bool
    isDeckSlotCompatible: bool | None = None
    magneticModuleEngageHeight: _NonNegativeNumber | None = None


class Parameters3(Parameters2, BaseModel):
    pass  # Currently equivalent to Parameters2.


class Dimensions(BaseModel):
    yDimension: _NonNegativeNumber
    zDimension: _NonNegativeNumber
    xDimension: _NonNegativeNumber


class _WellCommonMixin(BaseModel):
    model_config = ConfigDict(extra="allow")

    depth: _NonNegativeNumber
    totalLiquidVolume: _NonNegativeNumber
    x: _NonNegativeNumber
    z: _NonNegativeNumber
    geometryDefinitionId: str | None = None


class _WellCommon2(_WellCommonMixin):
    y: _NonNegativeNumber


class _WellCommon3(_WellCommonMixin):
    y: _NonPositiveNumber


class _CircularWellMixin(BaseModel):
    shape: Literal["circular"]
    diameter: _NonNegativeNumber


class _RectangularWellMixin(BaseModel):
    shape: Literal["rectangular"]
    xDimension: _NonNegativeNumber
    yDimension: _NonNegativeNumber


class CircularWellDefinition2(_WellCommon2, _CircularWellMixin):
    pass


class RectangularWellDefinition2(_WellCommon2, _RectangularWellMixin):
    pass


WellDefinition2 = Annotated[
    CircularWellDefinition2 | RectangularWellDefinition2, Discriminator("shape")
]


class CircularWellDefinition3(_WellCommon3, _CircularWellMixin):
    pass


class RectangularWellDefinition3(_WellCommon3, _RectangularWellMixin):
    pass


WellDefinition3 = Annotated[
    CircularWellDefinition3 | RectangularWellDefinition3, Discriminator("shape")
]


class SphericalSegment(BaseModel):
    shape: Spherical
    radiusOfCurvature: _NonNegativeNumber
    topHeight: _NonNegativeNumber
    bottomHeight: _NonNegativeNumber
    xCount: _StrictNonNegativeInt = 1
    yCount: _StrictNonNegativeInt = 1

    @cached_property
    def count(self) -> int:
        return self.xCount * self.yCount

    model_config = ConfigDict(ignored_types=(cached_property,))


class ConicalFrustum(BaseModel):
    shape: Conical
    bottomDiameter: _NonNegativeNumber
    topDiameter: _NonNegativeNumber
    topHeight: _NonNegativeNumber
    bottomHeight: _NonNegativeNumber
    xCount: _StrictNonNegativeInt = 1
    yCount: _StrictNonNegativeInt = 1

    def height_from_volume_search(self, target_volume: float) -> float:
        total_height = self.topHeight - self.bottomHeight
        max_height, min_height = total_height, 0.0
        volume_at_max_height = self.volume_from_height_circular(
            top_radius=self.topDiameter / 2,
            bottom_radius=self.bottomDiameter / 2,
            target_height=total_height,
            total_height=total_height,
        )
        if target_volume == volume_at_max_height:
            return max_height
        volume_at_min_height = self.volume_from_height_circular(
            top_radius=self.topDiameter / 2,
            bottom_radius=self.bottomDiameter / 2,
            target_height=0,
            total_height=total_height,
        )
        if target_volume == volume_at_min_height:
            return min_height

        y = total_height / 2
        volume_at_y = self.volume_from_height_circular(
            top_radius=self.topDiameter / 2,
            bottom_radius=self.bottomDiameter / 2,
            target_height=y,
            total_height=total_height,
        )
        guesses = [
            (volume_at_min_height, min_height),
            (volume_at_max_height, max_height),
        ]
        while abs(volume_at_y - target_volume) > RECURSIVE_SEARCH_VOLUME_TOLERANCE:
            max_height, max_volume = guesses[-1][1], guesses[-1][0]
            min_height, min_volume = guesses[0][1], guesses[0][0]

            # between volume_at_y and max value- undershot
            if volume_at_y < target_volume < max_volume:
                guesses = [(volume_at_y, y), (max_volume, max_height)]
            # overshot
            elif min_volume < target_volume < volume_at_y:
                guesses = [(min_volume, min_height), (volume_at_y, y)]
            y = (guesses[0][1] + guesses[1][1]) / 2

            volume_at_y = self.volume_from_height_circular(
                top_radius=self.topDiameter / 2,
                bottom_radius=self.bottomDiameter / 2,
                target_height=y,
                total_height=total_height,
            )
        return y

    def volume_from_height_circular(
        self,
        top_radius: float,
        bottom_radius: float,
        target_height: float,
        total_height: float,
    ) -> float:
        r_y = (target_height / total_height) * (
            top_radius - bottom_radius
        ) + bottom_radius
        return (pi * target_height / 3) * (
            bottom_radius**2 + bottom_radius * r_y + r_y**2
        )

    @cached_property
    def count(self) -> int:
        return self.xCount * self.yCount

    model_config = ConfigDict(ignored_types=(cached_property,))


class CuboidalFrustum(BaseModel):
    shape: Cuboidal
    bottomXDimension: _NonNegativeNumber
    bottomYDimension: _NonNegativeNumber
    topXDimension: _NonNegativeNumber
    topYDimension: _NonNegativeNumber
    topHeight: _NonNegativeNumber
    bottomHeight: _NonNegativeNumber
    xCount: _StrictNonNegativeInt = 1
    yCount: _StrictNonNegativeInt = 1

    @cached_property
    def count(self) -> int:
        return self.xCount * self.yCount

    model_config = ConfigDict(ignored_types=(cached_property,))


# A squared cone is the intersection of a cube and a cone that both
# share a central axis, and is a transitional shape between a cone and pyramid
"""
module RectangularPrismToCone(bottom_shape, diameter, x, y, z) {
    circle_radius = diameter/2;
    r1 = sqrt(x*x + y*y)/2;
    r2 = circle_radius/2;
    top_r = bottom_shape == "square" ? r1 : r2;
    bottom_r = bottom_shape == "square" ? r2 : r1;
    intersection() {
        cylinder(z,top_r,bottom_r,$fn=100);
        translate([0,0,z/2])cube([x, y, z], center=true);
    }
}
"""


class SquaredConeSegment(BaseModel):
    shape: SquaredCone
    bottomCrossSection: WellShape
    circleDiameter: _NonNegativeNumber
    rectangleXDimension: _NonNegativeNumber
    rectangleYDimension: _NonNegativeNumber
    topHeight: _NonNegativeNumber
    bottomHeight: _NonNegativeNumber
    xCount: _StrictNonNegativeInt = 1
    yCount: _StrictNonNegativeInt = 1

    @staticmethod
    def _area_trap_points(
        total_frustum_height: float,
        circle_diameter: float,
        rectangle_x: float,
        rectangle_y: float,
        dx: float,
    ) -> list[float]:
        """Grab a bunch of data points of area at given heights."""

        def _area_arcs(r: float, c: float, d: float) -> float:
            """Return the area of all 4 arc segments."""
            theata_y = asin(c / r)
            theata_x = asin(d / r)
            theata_arc = (pi / 2) - theata_y - theata_x
            # area of all 4 arcs is 4 * pi*r^2*(theata/2pi)
            return 2 * r**2 * theata_arc

        def _area(r: float) -> float:
            """Return the area of a given r_y."""
            # distance from the center of y axis of the rectangle to where the arc intercepts that side
            c: float = (
                sqrt(r**2 - (rectangle_y / 2) ** 2) if (rectangle_y / 2) < r else 0
            )
            # distance from the center of x axis of the rectangle to where the arc intercepts that side
            d: float = (
                sqrt(r**2 - (rectangle_x / 2) ** 2) if (rectangle_x / 2) < r else 0
            )
            arc_area = _area_arcs(r, c, d)
            y_triangles: float = rectangle_y * c
            x_triangles: float = rectangle_x * d
            return arc_area + y_triangles + x_triangles

        r_0 = circle_diameter / 2
        r_h = sqrt(rectangle_x**2 + rectangle_y**2) / 2

        num_steps = int(total_frustum_height / dx)
        points = [0.0]
        for i in range(num_steps + 1):
            r_y = (i * dx / total_frustum_height) * (r_h - r_0) + r_0
            points.append(_area(r_y))
        return points

    @cached_property
    def height_to_volume_table(self) -> dict[float, float]:
        """Return a lookup table of heights to volumes."""
        # the accuracy of this method is approximately +- 10*dx so for dx of 0.001 we have a +- 0.01 ul
        dx = 0.001
        total_height = self.topHeight - self.bottomHeight
        points = SquaredConeSegment._area_trap_points(
            total_height,
            self.circleDiameter,
            self.rectangleXDimension,
            self.rectangleYDimension,
            dx,
        )
        if self.bottomCrossSection is Rectangular:
            # The points function assumes the circle is at the bottom but if its flipped we just reverse the points
            points.reverse()
        elif self.bottomCrossSection is not Circular:
            raise NotImplementedError(
                "If you see this error a new well shape has been added without updating this code"
            )
        y = 0.0
        table: dict[float, float] = {}
        # fill in the table
        while y < total_height:
            table[y] = trapz(points[0 : int(y / dx)], dx=dx)
            y = y + dx

        # we always want to include the volume at the max height
        table[total_height] = trapz(points, dx=dx)
        return table

    @cached_property
    def volume_to_height_table(self) -> dict[float, float]:
        return dict((v, k) for k, v in self.height_to_volume_table.items())

    @cached_property
    def count(self) -> int:
        return self.xCount * self.yCount

    model_config = ConfigDict(ignored_types=(cached_property,))


"""
module filitedCuboidSquare(bottom_shape, diameter, width, length, height, steps) {
    module _slice(depth, x, y, r) {
        echo("called with: ", depth, x, y, r);
        circle_centers = [
            [(x/2)-r, (y/2)-r, 0],
            [(-x/2)+r, (y/2)-r, 0],
            [(x/2)-r, (-y/2)+r, 0],
            [(-x/2)+r, (-y/2)+r, 0]

        ];
        translate([0,0,depth/2])cube([x-2*r,y,depth], center=true);
        translate([0,0,depth/2])cube([x,y-2*r,depth], center=true);
        for (center = circle_centers) {
            translate(center) cylinder(depth, r, r, $fn=100);
        }
    }
    for (slice_height = [0:height/steps:height]) {
        r = (diameter) * (slice_height/height);
        translate([0,0,slice_height]) {
             _slice(height/steps , width, length, r/2);
        }
    }
}
module filitedCuboidForce(bottom_shape, diameter, width, length, height, steps) {
    module single_cone(r,x,y,z) {
        r = diameter/2;
        circle_face = [[ for (i = [0:1: steps]) i ]];
        theta = 360/steps;
        circle_points = [for (step = [0:1:steps]) [r*cos(theta*step), r*sin(theta*step), z]];
        final_points = [[x,y,0]];
        all_points = concat(circle_points, final_points);
        triangles = [for (step = [0:1:steps-1]) [step, step+1, steps+1]];
        faces = concat(circle_face, triangles);
        polyhedron(all_points, faces);
    }
    module square_section(r, x, y, z) {
        points = [
            [x,y,0],
            [-x,y,0],
            [-x,-y,0],
            [x,-y,0],
            [r,0,z],
            [0,r,z],
            [-r,0,z],
            [0,-r,z],
        ];
        faces = [
            [0,1,2,3],
            [4,5,6,7],
            [4, 0, 3],
            [5, 0, 1],
            [6, 1, 2],
            [7, 2, 3],
        ];
        polyhedron(points, faces);
    }
    circle_height = bottom_shape == "square" ? height : -height;
    translate_height = bottom_shape == "square" ? 0 : height;
    translate ([0,0, translate_height]) {
        union() {
            single_cone(diameter/2, width/2, length/2, circle_height);
            single_cone(diameter/2, -width/2, length/2, circle_height);
            single_cone(diameter/2, width/2, -length/2, circle_height);
            single_cone(diameter/2, -width/2, -length/2, circle_height);
            square_section(diameter/2, width/2, length/2, circle_height);
        }
    }
}

module filitedCuboid(bottom_shape, diameter, width, length, height) {
    if (width == length && width == diameter) {
        filitedCuboidSquare(bottom_shape, diameter, width, length, height, 100);
    }
    else {
        filitedCuboidForce(bottom_shape, diameter, width, length, height, 100);
    }
}"""


class RoundedCuboidSegment(BaseModel):
    shape: RoundedCuboid
    bottomCrossSection: WellShape
    circleDiameter: _NonNegativeNumber
    rectangleXDimension: _NonNegativeNumber
    rectangleYDimension: _NonNegativeNumber
    topHeight: _NonNegativeNumber
    bottomHeight: _NonNegativeNumber
    xCount: _StrictNonNegativeInt = 1
    yCount: _StrictNonNegativeInt = 1

    @cached_property
    def count(self) -> int:
        return self.xCount * self.yCount

    model_config = ConfigDict(ignored_types=(cached_property,))


class GroupMetadata(BaseModel):
    displayName: str | None = None
    displayCategory: DisplayCategory | None = None
    wellBottomShape: Literal["flat", "u", "v"] | None = None


class Group(BaseModel):
    wells: list[str]
    metadata: GroupMetadata
    brand: BrandData | None = None


WellSegment = Annotated[
    ConicalFrustum
    | CuboidalFrustum
    | SquaredConeSegment
    | RoundedCuboidSegment
    | SphericalSegment,
    Discriminator("shape"),
]


class HeightVolumePair(BaseModel):
    height: float
    volume: float


class InnerWellGeometry(BaseModel):
    sections: Annotated[list[WellSegment], Field(min_length=1)]


class UserDefinedVolumes(BaseModel):
    heightToVolumeMap: list[HeightVolumePair]


class Extents(BaseModel):
    total: AxisAlignedBoundingBox3D


class LabwareDefinition2(BaseModel):
    schemaVersion: Literal[2]
    version: Annotated[int, Field(ge=1)]
    namespace: Annotated[str, Field(pattern=SAFE_STRING_REGEX)]
    metadata: Metadata
    brand: BrandData
    parameters: Parameters2
    cornerOffsetFromSlot: Vector3D
    ordering: list[list[str]]
    dimensions: Dimensions
    wells: dict[str, WellDefinition2]
    groups: list[Group]
    stackingOffsetWithLabware: dict[str, Vector3D] = Field(default_factory=dict)
    stackingOffsetWithModule: dict[str, Vector3D] = Field(default_factory=dict)
    allowedRoles: list[LabwareRole] = Field(default_factory=list)
    gripperOffsets: dict[str, GripperOffsets] = Field(default_factory=dict)
    gripForce: float | None = None
    gripHeightFromLabwareBottom: float | None = None
    stackLimit: int | None = None
    compatibleParentLabware: list[str] | None = None
    innerLabwareGeometry: dict[
        str, InnerWellGeometry | UserDefinedVolumes
    ] | None = None


class LabwareDefinition3(BaseModel):
    otSharedSchema: Annotated[
        Literal["#/labware/schemas/3"], Field(alias="$otSharedSchema")
    ]
    schemaVersion: Literal[3]
    version: Annotated[int, Field(ge=1)]
    namespace: Annotated[str, Field(pattern=SAFE_STRING_REGEX)]
    metadata: Metadata
    brand: BrandData
    parameters: Parameters3
    ordering: list[list[str]]
    features: LocatingFeatures
    extents: Extents
    wells: dict[str, WellDefinition3]
    groups: list[Group]
    stackingOffsetWithLabware: dict[str, Vector3D] = Field(default_factory=dict)
    stackingOffsetWithModule: dict[str, Vector3D] = Field(default_factory=dict)
    allowedRoles: list[LabwareRole] = Field(default_factory=list)
    gripperOffsets: dict[str, GripperOffsets] = Field(default_factory=dict)
    gripForce: float | None = None
    gripHeightFromLabwareOrigin: float | None = None
    stackLimit: int | None = None
    compatibleParentLabware: list[str] | None = None
    innerLabwareGeometry: dict[str, InnerWellGeometry] | None = None


LabwareDefinition = Annotated[
    LabwareDefinition2 | LabwareDefinition3,
    Discriminator("schemaVersion"),
]


labware_definition_type_adapter: Final = TypeAdapter[LabwareDefinition](
    LabwareDefinition
)



from typing import List
from pathlib import Path


def list_fixtures(version: int) -> List[Path]:
    base = Path(__file__).parent / ".." / ".." / "protocol" / "fixtures" / f"{version}"
    return list(base.iterdir())



import json
import pytest
from typing import Any, Dict
from pathlib import Path
from opentrons_shared_data import load_shared_data
from opentrons_shared_data.protocol.models import (
    protocol_schema_v6,
    Robot,
    Metadata,
    Liquid,
    Module,
    Pipette,
    Labware,
)

from . import list_fixtures


@pytest.mark.parametrize("defpath", list_fixtures(6))
def test_v6_types(defpath: Path) -> None:
    def_data = load_shared_data(defpath)
    def_model = protocol_schema_v6.ProtocolSchemaV6.model_validate_json(def_data)
    def_dict_from_model = def_model.model_dump(
        exclude_unset=True,
        # 'schemaVersion' in python is '$schemaVersion' in JSON
        by_alias=True,
    )
    expected_def_dict = json.loads(def_data)
    delete_unexpected_results(expected_def_dict)
    assert def_dict_from_model == expected_def_dict


# TODO (Tamar and Max 3/16/22):
# Some of our v6 fixtures accidentally contain command "result" data,
# which is not part of the schema, and which we don't parse into our models.
# Here, we delete the extra junk from the fixtures.
# Remove this when the fixtures are corrected.
# https://github.com/Opentrons/opentrons/issues/9701
def delete_unexpected_results(protocol_fixture: Dict[str, Any]) -> None:
    for command_object_dict in protocol_fixture["commands"]:
        command_object_dict.pop("result", None)
        command_object_dict.pop("id", None)


@pytest.mark.parametrize(
    "input_command, missing_id, load_command",
    [
        (
            protocol_schema_v6.Command(
                commandType="loadLabware",
                params=protocol_schema_v6.Params(labwareId="labware-id-3"),
            ),
            "labware-id-3",
            "loadLabware",
        ),
        (
            protocol_schema_v6.Command(
                commandType="loadPipette",
                params=protocol_schema_v6.Params(pipetteId="pipette-id-3"),
            ),
            "pipette-id-3",
            "loadPipette",
        ),
        (
            protocol_schema_v6.Command(
                commandType="loadLiquid",
                params=protocol_schema_v6.Params(liquidId="liquid-id-3"),
            ),
            "liquid-id-3",
            "loadLiquid",
        ),
        (
            protocol_schema_v6.Command(
                commandType="loadModule",
                params=protocol_schema_v6.Params(moduleId="module-id-3"),
            ),
            "module-id-3",
            "loadModule",
        ),
    ],
)
def test_schema_validators(
    input_command: protocol_schema_v6.Command, missing_id: str, load_command: str
) -> None:
    """Should raise an error the keys do not match."""
    labware = {
        "labware-id-1": Labware(definitionId="definition-1"),
        "labware-id-2": Labware(definitionId="definition-2"),
    }
    pipettes = {"pipette-id-1": Pipette(name="pipette-1")}
    liquids = {"liquid-id-1": Liquid(displayName="liquid-1", description="liquid desc")}
    modules = {"module-id-1": Module(model="model-1")}
    with pytest.raises(
        ValueError,
        match=f"{load_command} command at index 0 references ID {missing_id}, which doesn't exist.",
    ):
        protocol_schema_v6.ProtocolSchemaV6(
            otSharedSchema="#/protocol/schemas/6",
            schemaVersion=6,
            metadata=Metadata(),
            robot=Robot(model="OT-2 Standard", deckId=""),
            labware=labware,
            pipettes=pipettes,
            liquids=liquids,
            modules=modules,
            labwareDefinitions={},
            commands=[input_command],
        )



"""Error codes."""

from enum import Enum
from dataclasses import dataclass
import json
from typing import Tuple, Dict
from functools import lru_cache

from ..load import load_shared_data
from .categories import ErrorCategories

CODES: Dict[str, Dict[str, str]] = json.loads(
    load_shared_data("errors/definitions/1/errors.json")
)["codes"]


@dataclass(frozen=True)
class ErrorCode:
    """Individual error code data."""

    code: str
    detail: str
    category: ErrorCategories


def _code_from_dict_entry(entry_code: str) -> ErrorCode:
    return ErrorCode(
        code=entry_code,
        detail=CODES[entry_code]["detail"],
        category=ErrorCategories.by_category_name(CODES[entry_code]["category"]),
    )


class ErrorCodes(Enum):
    """All enumerated error codes."""

    COMMUNICATION_ERROR = _code_from_dict_entry("1000")
    CANBUS_COMMUNICATION_ERROR = _code_from_dict_entry("1001")
    INTERNAL_USB_COMMUNICATION_ERROR = _code_from_dict_entry("1002")
    MODULE_COMMUNICATION_ERROR = _code_from_dict_entry("1003")
    COMMAND_TIMED_OUT = _code_from_dict_entry("1004")
    FIRMWARE_UPDATE_FAILED = _code_from_dict_entry("1005")
    INTERNAL_MESSAGE_FORMAT_ERROR = _code_from_dict_entry("1006")
    CANBUS_CONFIGURATION_ERROR = _code_from_dict_entry("1007")
    CANBUS_BUS_ERROR = _code_from_dict_entry("1008")
    ROBOTICS_CONTROL_ERROR = _code_from_dict_entry("2000")
    MOTION_FAILED = _code_from_dict_entry("2001")
    HOMING_FAILED = _code_from_dict_entry("2002")
    STALL_OR_COLLISION_DETECTED = _code_from_dict_entry("2003")
    MOTION_PLANNING_FAILURE = _code_from_dict_entry("2004")
    POSITION_ESTIMATION_INVALID = _code_from_dict_entry("2005")
    MOVE_CONDITION_NOT_MET = _code_from_dict_entry("2006")
    CALIBRATION_STRUCTURE_NOT_FOUND = _code_from_dict_entry("2007")
    EDGE_NOT_FOUND = _code_from_dict_entry("2008")
    EARLY_CAPACITIVE_SENSE_TRIGGER = _code_from_dict_entry("2009")
    INACCURATE_NON_CONTACT_SWEEP = _code_from_dict_entry("2010")
    MISALIGNED_GANTRY = _code_from_dict_entry("2011")
    UNMATCHED_TIP_PRESENCE_STATES = _code_from_dict_entry("2012")
    POSITION_UNKNOWN = _code_from_dict_entry("2013")
    EXECUTION_CANCELLED = _code_from_dict_entry("2014")
    FAILED_GRIPPER_PICKUP_ERROR = _code_from_dict_entry("2015")
    MOTOR_DRIVER_ERROR = _code_from_dict_entry("2016")
    PIPETTE_LIQUID_NOT_FOUND = _code_from_dict_entry("2017")
    TIP_HIT_WELL_BOTTOM = _code_from_dict_entry("2018")
    STACKER_STALL_OR_COLLISION_DETECTED = _code_from_dict_entry("2019")
    ROBOTICS_INTERACTION_ERROR = _code_from_dict_entry("3000")
    LABWARE_DROPPED = _code_from_dict_entry("3001")
    LABWARE_NOT_PICKED_UP = _code_from_dict_entry("3002")
    TIP_PICKUP_FAILED = _code_from_dict_entry("3003")
    TIP_DROP_FAILED = _code_from_dict_entry("3004")
    UNEXPECTED_TIP_REMOVAL = _code_from_dict_entry("3005")
    PIPETTE_OVERPRESSURE = _code_from_dict_entry("3006")
    E_STOP_ACTIVATED = _code_from_dict_entry("3008")
    E_STOP_NOT_PRESENT = _code_from_dict_entry("3009")
    PIPETTE_NOT_PRESENT = _code_from_dict_entry("3010")
    GRIPPER_NOT_PRESENT = _code_from_dict_entry("3011")
    UNEXPECTED_TIP_ATTACH = _code_from_dict_entry("3012")
    FIRMWARE_UPDATE_REQUIRED = _code_from_dict_entry("3013")
    INVALID_ACTUATOR = _code_from_dict_entry("3014")
    MODULE_NOT_PRESENT = _code_from_dict_entry("3015")
    INVALID_INSTRUMENT_DATA = _code_from_dict_entry("3016")
    INVALID_LIQUID_CLASS_NAME = _code_from_dict_entry("3017")
    TIP_DETECTOR_NOT_FOUND = _code_from_dict_entry("3018")
    HEPA_UV_FAILED = _code_from_dict_entry("3019")
    STACKER_SHUTTLE_MISSING = _code_from_dict_entry("3020")
    STACKER_SHUTTLE_LABWARE_FAILED = _code_from_dict_entry("3021")
    STACKER_HOPPER_LABWARE_FAILED = _code_from_dict_entry("3022")
    STACKER_SHUTTLE_OCCUPIED = _code_from_dict_entry("3023")
    GENERAL_ERROR = _code_from_dict_entry("4000")
    ROBOT_IN_USE = _code_from_dict_entry("4001")
    API_REMOVED = _code_from_dict_entry("4002")
    NOT_SUPPORTED_ON_ROBOT_TYPE = _code_from_dict_entry("4003")
    COMMAND_PRECONDITION_VIOLATED = _code_from_dict_entry("4004")
    COMMAND_PARAMETER_LIMIT_VIOLATED = _code_from_dict_entry("4005")
    INVALID_PROTOCOL_DATA = _code_from_dict_entry("4006")
    API_MISCONFIGURATION = _code_from_dict_entry("4007")
    INVALID_STORED_DATA = _code_from_dict_entry("4008")
    MISSING_CONFIGURATION_DATA = _code_from_dict_entry("4009")
    RUNTIME_PARAMETER_VALUE_REQUIRED = _code_from_dict_entry("4010")
    INCORRECT_API_VERSION = _code_from_dict_entry("4011")

    @classmethod
    @lru_cache(25)
    def by_error_code(cls, error_code: str) -> "ErrorCodes":
        """Get an error by its code.

        Note: this is a linear search because it makes life easier and also this is rare in python.
        """
        for error in cls:
            if error.value.code == error_code:
                return error
        raise KeyError(error_code)

    @classmethod
    @lru_cache(len(ErrorCategories))
    def of_category(cls, category: ErrorCategories) -> Tuple["ErrorCodes", ...]:
        """Get all error codes by their category."""
        return tuple(
            error
            for error in cls
            if error.value.code.startswith(category.value.code_prefix)
        )



import argparse
import csv
import json
from ast import literal_eval
from typing import Optional, Dict, Union, Any, List
from pathlib import Path
from pydantic import BaseModel
import math

from ... import get_shared_data_root
from ..pipette_definition import (
    PipetteGeometryDefinition,
    PipetteLiquidPropertiesDefinition,
    PipettePhysicalPropertiesDefinition,
    PlungerPositions,
    SupportedTipsDefinition,
    MotorConfigurations,
    PartialTipDefinition,
    AvailableSensorDefinition,
    PickUpTipConfigurations,
    PressFitPickUpTipConfiguration,
    DropTipConfigurations,
    PlungerEjectDropTipConfiguration,
)

from ..types import PipetteModelSpec


PIPETTE_DEFINITION_ROOT = Path("pipette") / "definitions" / "2"
GEOMETRY_ROOT = get_shared_data_root() / PIPETTE_DEFINITION_ROOT / "geometry"
GENERAL_ROOT = get_shared_data_root() / PIPETTE_DEFINITION_ROOT / "general"
LIQUID_ROOT = get_shared_data_root() / PIPETTE_DEFINITION_ROOT / "liquid"

GENERAL_SCHEMA = "#/pipette/schemas/2/pipettePropertiesSchema.json"
LIQUID_SCHEMA = "#/pipette/schemas/2/pipetteLiquidPropertiesSchema.json"
GEOMETRY_SCHEMA = "#/pipette/schemas/2/pipetteGeometryPropertiesSchema.json"


def _build_pickup_tip_data(
    model_configurations: Optional[PipetteModelSpec] = None,
) -> PickUpTipConfigurations:
    presses = 0
    increment = 0
    distance = 0.0
    if model_configurations:
        current = model_configurations["pickUpCurrent"]["value"]
        speed = model_configurations["pickUpSpeed"]["value"]
        presses = model_configurations["pickUpPresses"]["value"]
        increment = int(model_configurations["pickUpIncrement"]["value"])
        distance = model_configurations["pickUpDistance"]["value"]
    else:
        print("Handling pick up tip configurations\n")
        speed = float(input("please provide the speed\n"))
        presses = int(input("please provide the number of presses for force pick up\n"))
        increment = int(
            input(
                "please provide the increment to move the pipette down for force pickup\n"
            )
        )
        distance = float(
            input("please provide the starting distance for pick up tip\n")
        )
    print(
        f"TODO: Current {current}, speed {speed} and distance {distance} is not used yet"
    )
    return PickUpTipConfigurations(
        pressFit=PressFitPickUpTipConfiguration(
            presses=presses, increment=increment, configurationsByNozzleMap={}
        )
    )


def _build_drop_tip_data(
    model_configurations: Optional[PipetteModelSpec] = None,
) -> DropTipConfigurations:
    if model_configurations:
        current = model_configurations["dropTipCurrent"]["value"]
        speed = model_configurations["dropTipSpeed"]["value"]
    else:
        print("Handling drop tip configurations\n")
        speed = float(input("please provide the speed\n"))
    return DropTipConfigurations(
        plungerEject=PlungerEjectDropTipConfiguration(
            current=current,
            speed=speed,
        ),
        camAction=None,
    )


def _build_plunger_positions(
    model_configurations: Optional[PipetteModelSpec] = None,
) -> PlungerPositions:
    if model_configurations:
        top = model_configurations["top"]["value"]
        bottom = model_configurations["bottom"]["value"]
        blow_out = model_configurations["blowout"]["value"]
        drop_tip = model_configurations["dropTip"]["value"]
    else:
        print("Handling plunger motor positions\n")
        top = float(input("Please provide the top plunger position\n"))
        bottom = float(input("Please provide the bottom plunger position\n"))
        blow_out = float(input("Please provide the blow out plunger position\n"))
        drop_tip = float(input("Please provide the drop tip plunger position\n"))
    return PlungerPositions(top=top, bottom=bottom, blowout=blow_out, drop=drop_tip)


def _build_motor_configurations(
    model_configurations: Optional[PipetteModelSpec] = None,
) -> MotorConfigurations:
    if model_configurations:
        run = model_configurations["plungerCurrent"]["value"]
        idle = model_configurations.get("idleCurrent", run)
    else:
        print("Handling default plunger motor currents\n")
        idle = float(
            input("Please provide the default idle current of the plunger motor\n")
        )
        run = float(
            input("Please provide the default run current of the plunger motor\n")
        )
    return MotorConfigurations(idle=idle, run=run)


def _build_partial_tip_configurations(channels: int) -> PartialTipDefinition:
    if channels == 8:
        return PartialTipDefinition(
            partialTipSupported=True, availableConfigurations=[1, 2, 3, 4, 5, 6, 7, 8]
        )
    elif channels == 96:
        return PartialTipDefinition(
            partialTipSupported=True, availableConfigurations=[1, 8, 12, 96]
        )
    else:
        return PartialTipDefinition(partialTipSupported=False)


def build_geometry_model_v2(
    input_dictionary: Dict[str, Any]
) -> PipetteGeometryDefinition:
    return PipetteGeometryDefinition.model_validate(input_dictionary)


def build_liquid_model_v2(
    input_dictionary: Dict[str, Any],
    supported_tip_configurations: Dict[str, SupportedTipsDefinition],
) -> PipetteLiquidPropertiesDefinition:
    if input_dictionary:
        if input_dictionary.get("partialTipConfigurations"):
            return PipetteLiquidPropertiesDefinition.model_validate(
                {**input_dictionary, "supportedTips": supported_tip_configurations}
            )
        else:
            return PipetteLiquidPropertiesDefinition.model_validate(
                {
                    **input_dictionary,
                    "supportedTips": supported_tip_configurations,
                }
            )
    max_volume = int(input("please provide the max volume of the pipette\n"))
    min_volume = float(input("please provide the min volume of the pipette\n"))
    default_tipracks = input(
        "please input the load names of default tipracks separated by commas\n"
    )
    list_default_tipracks = default_tipracks.split(",")
    return PipetteLiquidPropertiesDefinition.model_validate(
        {
            "supportedTips": supported_tip_configurations,
            "maxVolume": max_volume,
            "minVolume": min_volume,
            "defaultTipracks": list_default_tipracks,
        }
    )


def build_physical_model_v2(
    input_dictionary: Optional[Dict[str, Any]], pipette_type: Optional[str] = None
) -> PipettePhysicalPropertiesDefinition:
    if input_dictionary:
        available_sensors = AvailableSensorDefinition(
            sensors=input_dictionary.pop("availableSensors", [])
        )
        back_compat_names = input_dictionary.pop("backCompatNames", [])
        return PipettePhysicalPropertiesDefinition.model_validate(
            {
                **input_dictionary,
                "availableSensors": available_sensors,
                "backCompatNames": back_compat_names,
            }
        )
    print(f"Handling general pipette information for {pipette_type}\n")
    display_name = input("please provide the product name of the pipette\n")

    display_category = input("please provide the model generation of the pipette\n")
    input_available_sensors = input(
        "Please provide a list of available sensors, separated by comma\n"
    )
    channels = input(f"Please provide the number of channels your {pipette_type} has\n")
    shaft_diam = float(input(f"Please provide the shaft diameter of {pipette_type}\n"))
    shaft_ul_per_mm = float(
        input(f"Please provide the uL to mm conversion for {pipette_type}\n")
    )
    pick_up_tip_configurations = _build_pickup_tip_data()
    drop_tip_configurations = _build_drop_tip_data()
    plunger_positions = _build_plunger_positions()
    plunger_motor_configurations = _build_motor_configurations()
    partial_tip_configurations = _build_partial_tip_configurations(int(channels))

    back_compat_names_str = input(
        "Please list compatible pipette names separated by commas or hit enter if none"
    )
    if back_compat_names_str:
        back_compat_names = [i.strip() for i in back_compat_names_str.split(",")]
    else:
        back_compat_names = []
    return PipettePhysicalPropertiesDefinition.model_validate(
        {
            "displayName": display_name,
            "model": pipette_type,
            "displayCategory": display_category,
            "pickUpTipConfigurations": pick_up_tip_configurations,
            "dropTipConfigurations": drop_tip_configurations,
            "plungerMotorConfigurations": plunger_motor_configurations,
            "plungerPositionsConfigurations": plunger_positions,
            "availableSensors": AvailableSensorDefinition(
                sensors=input_available_sensors.split(",")
            ),
            "partialTipConfigurations": partial_tip_configurations,
            "channels": channels,
            "shaftDiameter": shaft_diam,
            "shaftULperMM": shaft_ul_per_mm,
            "backCompatNames": back_compat_names,
        }
    )


def build_supported_tips(input_dictionary: Dict[str, Any]) -> SupportedTipsDefinition:
    return SupportedTipsDefinition.model_validate(input_dictionary)


def save_to_file(
    directorypath: Path,
    file_name: str,
    data: Union[BaseModel, Dict[str, Any]],
    schema_path: str,
) -> None:
    """
    Function used to save data to a file
    """

    directorypath.mkdir(parents=True, exist_ok=True)
    filepath = directorypath / f"{file_name}.json"
    if isinstance(data, BaseModel):
        # We have to rely on the json serialization of the
        # pydantic model to properly convert special objects
        # such as (pipetteModelType) to a string. However,
        # we also want to add in a new key on top of the
        # data in the pydantic model so we need to convert it
        # back to a dictionary before saving to file.
        json_basemodel = data.json(by_alias=True)
        dict_basemodel = json.loads(json_basemodel)
        dict_basemodel["$otSharedSchema"] = schema_path
        filepath.write_text(
            json.dumps(dict_basemodel, ensure_ascii=False), encoding="utf-8"
        )
    else:
        data["$otSharedSchema"] = schema_path
        filepath.write_text(json.dumps(data, ensure_ascii=False), encoding="utf-8")


def migrate_new_blow_out_configs_v2() -> None:
    pipette_volumes = {
        1: ["p10", "p50", "p300", "p1000"],
        2: ["p20", "p300", "p1000"],
        3: ["p50", "p1000"],
    }

    for pipette_gen in range(1, 4):  # gen 1, gen 2, gen 3
        print(f"\nTaking data for gen{pipette_gen} pipettes:")
        # overwrite and pass in dict with user input for each gen
        shaft_diameters = {
            vol: float(input(f"Enter shaft diameter for gen{pipette_gen} {vol}: "))
            for vol in pipette_volumes[pipette_gen]
        }
        fill_blowout_configs(pipette_gen, shaft_diameters, pipette_volumes[pipette_gen])


def fill_blowout_configs(
    pipette_gen: int, shaft_diameters: Dict[str, float], volumes: List[str]
) -> None:

    general_config_files = Path(GENERAL_ROOT).glob("*")
    for pipette_type in general_config_files:  # single, eight, 96-channel
        for volume in volumes:  # pipette max volume- p10, p20, p50, etc.
            shaft_diameter = shaft_diameters[volume]
            # calculate uL per mm, default blowout vol
            ul_per_mm = math.pi * (shaft_diameter / 2) ** 2
            # get path for every file version for the pipette gen at hand
            volume_path = Path(pipette_type / volume)
            for general_file_version in volume_path.glob(f"{str(pipette_gen)}_*.json"):
                with open(general_file_version, "r") as file:
                    general_config_dict = json.load(file)
                general_config_dict["shaftDiameter"] = round(shaft_diameter, 3)
                general_config_dict["shaftULperMM"] = round(ul_per_mm, 3)

                # get file path to pass into save_to_file
                gen_path = GENERAL_ROOT / pipette_type / volume
                version_str = str(general_file_version).split("/")[-1][:-5]
                save_to_file(gen_path, version_str, general_config_dict, GENERAL_SCHEMA)


def build_new_pipette_model_v2(
    pipette_functions_dict: Dict[str, Any], pipette_model_csv: str
) -> None:
    """
    Build a brand new pipette model from user inputted data.
    """
    top_level_pipette_model = {}
    with open(pipette_model_csv, "r") as f:
        reader = csv.reader(f)
        current_key = None
        for row in reader:
            if row[0]:
                current_key = row[0]
                top_level_pipette_model[current_key] = {row[1]: literal_eval(row[2])}
            elif current_key:
                top_level_pipette_model[current_key][row[1]] = literal_eval(row[2])
            else:
                continue
    geometry_model = build_geometry_model_v2(top_level_pipette_model["geometry"])
    liquid_model = build_liquid_model_v2(
        top_level_pipette_model["liquid"],
        pipette_functions_dict,
    )
    liquid_model_dict = liquid_model.model_dump(by_alias=True)
    liquid_model_dict["supportedTips"] = {
        k.name: v for k, v in liquid_model_dict["supportedTips"].items()
    }
    pipette_type = f"p{liquid_model.max_volume}"
    physical_model = build_physical_model_v2(
        {
            **top_level_pipette_model["general"],
            "plungerMotorConfigurations": top_level_pipette_model[
                "plungerMotorConfigurations"
            ],
            "plungerPositionsConfigurations": top_level_pipette_model[
                "plungerPositionsConfigurations"
            ],
            "pickUpTipConfigurations": top_level_pipette_model[
                "pickUpTipConfigurations"
            ],
            "dropTipConfigurations": top_level_pipette_model["dropTipConfigurations"],
            "partialTipConfigurations": _build_partial_tip_configurations(
                top_level_pipette_model["general"]["channels"]
            ),
        },
        pipette_type,
    )

    major_version = input("please provide the major version of this pipette model\n")
    minor_version = input("please provide the minor version of this pipette model\n")

    file_name = f"{major_version}_{minor_version}"
    current_pipette_path = Path(physical_model.channels.name.lower()) / pipette_type

    save_to_file(
        GEOMETRY_ROOT / current_pipette_path, file_name, geometry_model, GEOMETRY_SCHEMA
    )
    save_to_file(
        GENERAL_ROOT / current_pipette_path, file_name, physical_model, GENERAL_SCHEMA
    )
    save_to_file(
        LIQUID_ROOT / current_pipette_path, file_name, liquid_model_dict, LIQUID_SCHEMA
    )


def main() -> None:
    """Entry point."""
    parser = argparse.ArgumentParser(
        description="96 channel tip handling testing script."
    )
    parser.add_argument(
        "--path_to_pipette_model",
        type=str,
        help="the csv filled with data to build a pipette model",
        default=None,
    )
    parser.add_argument(
        "--migrate_blowout_configs",
        type=bool,
        help="If true, migrate new blowout configs to existing json files",
        default=False,
    )

    args = parser.parse_args()
    if args.migrate_blowout_configs:
        migrate_new_blow_out_configs_v2()
    else:
        using_csv = "" if args.path_to_pipette_model else "out"
        print(f"Building new pipette function with{using_csv} using csv")

        next_tip = True
        pipette_functions_dict: Dict[str, Any] = {}
        while next_tip:
            print("Preparing pipetting function table from csv\n")
            tip_volume = input("Please provide the next tip volume\n")
            tiprack = input(
                "Please provide the tiprack/tip type associated with this data\n"
            )
            path_to_data = input("Please provide the data path to the csv\n")
            csv_dict = pipette_functions_dict.get(f"t{tip_volume}", {})
            with open(path_to_data, "r") as f:
                reader = csv.reader(f)
                pipetting_key = None
                for row in reader:
                    current_key_exists = csv_dict.get(row[0])
                    if row[0] == "aspirate" or row[0] == "dispense":
                        pipetting_key = row[0]
                        if current_key_exists:
                            csv_dict[pipetting_key][f"{tiprack}"].append(
                                [float(row[1]), float(row[2]), float(row[3])]
                            )
                        else:
                            csv_dict[pipetting_key] = {
                                f"{tiprack}": [
                                    [float(row[1]), float(row[2]), float(row[3])]
                                ]
                            }
                    elif pipetting_key and not row[0]:
                        csv_dict[pipetting_key][f"{tiprack}"].append(
                            [float(row[1]), float(row[2]), float(row[3])]
                        )
                    else:
                        pipetting_key = None
                        if current_key_exists:
                            continue
                        csv_dict[row[0]] = row[1]
            pipette_functions_dict[f"t{tip_volume}"] = csv_dict
            cont = input("Additional tips? Please enter y/n")

            next_tip = cont == "y" or cont == "yes"
        converted_pipette_functions_dict = {
            k: build_supported_tips(d) for k, d in pipette_functions_dict.items()
        }

        print(
            "Please review the aspirate and dispense functions once the pipette model is built.\n"
        )
        build_new_pipette_model_v2(
            converted_pipette_functions_dict, args.path_to_pipette_model
        )


if __name__ == "__main__":
    """
    A script to automate building a pipette configuration definition.

    This script can either perform migrations from a v1 -> v2 schema format
    or build a brand new script from scratch.

    When building a new pipette configuration model, you will either need
    to provide CSVs or use command line inputs.

    If you choose CSVs you will need one CSV for the general pipette configuration
    data (such as pipette model or number of channels) and one for every tip
    type that this pipette model can support.
    """
    main()



from typing import List, Tuple

from pathlib import Path


def get_ot_defs(schema: int) -> List[Tuple[str, int]]:
    def_files = (
        Path(__file__).parent / ".." / ".." / "labware" / "definitions" / str(schema)
    ).glob("**/*.json")

    # example filename
    # shared-data/labware/definitions/2/opentrons_96_tiprack_300ul/1.json
    return [(f.parent.name, int(f.stem)) for f in def_files]



"""Python shared data models for liquid class definitions."""

from enum import Enum
from typing import Literal, Union, Optional, Sequence, Tuple, Any, Type

from pydantic import (
    ConfigDict,
    BaseModel,
    field_validator,
    ValidationInfo,
    Field,
    StrictInt,
    StrictFloat,
    StrictBool,
    model_validator,
)
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Annotated


_StrictNonNegativeInt = Annotated[int, Field(strict=True, ge=0)]
_StrictGreaterThanZeroInt = Annotated[int, Field(strict=True, gt=0)]
_StrictGreaterThanZeroFloat = Annotated[float, Field(strict=True, gt=0.0)]
_StrictNonNegativeFloat = Annotated[float, Field(strict=True, ge=0.0)]


_Number = Union[StrictInt, StrictFloat]
"""JSON number type, written to preserve lack of decimal point"""

_NonNegativeNumber = Union[_StrictNonNegativeInt, _StrictNonNegativeFloat]
"""Non-negative JSON number type, written to preserve lack of decimal point."""

_GreaterThanZeroNumber = Union[_StrictGreaterThanZeroInt, _StrictGreaterThanZeroFloat]

LiquidHandlingPropertyByVolume = Sequence[Tuple[_NonNegativeNumber, _NonNegativeNumber]]
"""Settings for liquid class settings that are interpolated by volume."""

CorrectionByVolume = Sequence[Tuple[_NonNegativeNumber, _Number]]
"""Settings for correctionByVolume, which unlike other `byVolume` properties allows negative values with volume."""


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default")


# Make sure these always match the values in the Enums below
POSITION_REFERENCE_VALUE_TYPE = Literal[
    "well-bottom", "well-top", "well-center", "liquid-meniscus"
]
"""Values for positionReference."""

BLOWOUT_LOCATION_VALUE_TYPE = Literal["source", "destination", "trash"]


# Make sure the values of these Enums always match the Literals above
class PositionReference(Enum):
    """Positional reference for liquid handling operations."""

    WELL_BOTTOM = "well-bottom"
    WELL_TOP = "well-top"
    WELL_CENTER = "well-center"
    LIQUID_MENISCUS = "liquid-meniscus"


class BlowoutLocation(Enum):
    """Location for blowout during a transfer function."""

    SOURCE = "source"
    DESTINATION = "destination"
    TRASH = "trash"


class Coordinate(BaseModel):
    """Three-dimensional coordinates."""

    x: _Number
    y: _Number
    z: _Number


class BaseLiquidClassModel(BaseModel):
    """Base class for liquid class definitions."""

    model_config = ConfigDict(populate_by_name=True, extra="forbid")


class TipPosition(BaseLiquidClassModel):
    """Properties for tip position reference and relative offset."""

    positionReference: PositionReference = Field(
        ...,
        alias="position_reference",
        description="Position reference for tip position.",
    )
    offset: Coordinate = Field(
        ..., description="Relative offset from position reference."
    )


class DelayParams(BaseLiquidClassModel):
    """Parameters for delay."""

    duration: _NonNegativeNumber = Field(
        ..., description="Duration of delay, in seconds."
    )


class DelayProperties(BaseLiquidClassModel):
    """Shared properties for delay."""

    enable: StrictBool = Field(..., description="Whether delay is enabled.")
    params: DelayParams | SkipJsonSchema[None] = Field(
        None,
        description="Parameters for the delay function.",
        json_schema_extra=_remove_default,
    )

    @model_validator(mode="before")
    @classmethod
    def reshape(cls, data: Any) -> Any:
        """Move any params specified as top-level keys into the 'params' value."""
        if isinstance(data, dict):
            if None not in (data.get("enable"), data.get("enabled")):
                raise ValueError(
                    "Delay properties should specify either 'enable' or 'enabled', not both."
                )
            if data.get("enabled") is not None:
                data["enable"] = data["enabled"]
                data.pop("enabled")
            if "duration" in data.keys():
                if data.get("params"):
                    raise ValueError(
                        "Delay properties should specify either duration or params, not both."
                    )
                data["params"] = DelayParams(duration=data["duration"])
                data.pop("duration")
        return data

    @field_validator("params")
    @classmethod
    def _validate_params(
        cls, v: Optional[DelayParams], info: ValidationInfo
    ) -> Optional[DelayParams]:
        if v is None and info.data.get("enable", False):
            raise ValueError("If enable is true parameters for delay must be defined.")
        return v


class LiquidClassTouchTipParams(BaseLiquidClassModel):
    """Parameters for touch-tip."""

    # Note: Do not call this `TouchTipParams`, because that class name is used by the
    # unrelated touchTip command in PE. Both classes are exported to things like the
    # command schema JSON files, so the classes can't have the same name.

    zOffset: _Number = Field(
        ...,
        alias="z_offset",
        description="Offset from the top of the well for touch-tip, in millimeters.",
    )
    mmFromEdge: _Number = Field(
        ...,
        alias="mm_from_edge",
        description="Offset away from the the well edge, in millimeters.",
    )
    speed: _GreaterThanZeroNumber = Field(
        ..., alias="speed", description="Touch-tip speed, in millimeters per second."
    )


class TouchTipProperties(BaseLiquidClassModel):
    """Shared properties for the touch-tip function."""

    enable: StrictBool = Field(..., description="Whether touch-tip is enabled.")
    params: LiquidClassTouchTipParams | SkipJsonSchema[None] = Field(
        None,
        description="Parameters for the touch-tip function.",
        json_schema_extra=_remove_default,
    )

    @model_validator(mode="before")
    @classmethod
    def reshape(cls, data: Any) -> Any:
        """Move any params specified as top-level keys into the 'params' value."""
        return reshape_glob(
            data=data,
            params_model=LiquidClassTouchTipParams,
            property_name="Touch tip properties",
        )

    @field_validator("params")
    @classmethod
    def _validate_params(
        cls, v: Optional[LiquidClassTouchTipParams], info: ValidationInfo
    ) -> Optional[LiquidClassTouchTipParams]:
        if v is None and info.data.get("enable", False):
            raise ValueError(
                "If enable is true parameters for touch tip must be defined."
            )
        return v


class MixParams(BaseLiquidClassModel):
    """Parameters for mix."""

    repetitions: _StrictNonNegativeInt = Field(
        ...,
        alias="repetitions",
        description="Number of mixing repetitions. 0 is valid, but no mixing will occur.",
    )
    volume: _GreaterThanZeroNumber = Field(
        ..., alias="volume", description="Volume used for mixing, in microliters."
    )


class MixProperties(BaseLiquidClassModel):
    """Mixing properties."""

    enable: StrictBool = Field(..., description="Whether mix is enabled.")
    params: MixParams | SkipJsonSchema[None] = Field(
        None,
        description="Parameters for the mix function.",
        json_schema_extra=_remove_default,
    )

    @model_validator(mode="before")
    @classmethod
    def reshape(cls, data: Any) -> Any:
        """Move any params specified as top-level keys into the 'params' value."""
        return reshape_glob(
            data=data, params_model=MixParams, property_name="Mix properties"
        )

    @field_validator("params")
    @classmethod
    def _validate_params(
        cls, v: Optional[MixParams], info: ValidationInfo
    ) -> Optional[MixParams]:
        if v is None and info.data.get("enable", False):
            raise ValueError("If enable is true parameters for mix must be defined.")
        return v


class BlowoutParams(BaseLiquidClassModel):
    """Parameters for blowout."""

    location: BlowoutLocation = Field(
        ..., alias="location", description="Location well or trash entity for blow out."
    )
    flowRate: _GreaterThanZeroNumber = Field(
        ...,
        alias="flow_rate",
        description="Flow rate for blow out, in microliters per second.",
    )


class BlowoutProperties(BaseLiquidClassModel):
    """Blowout properties."""

    enable: StrictBool = Field(..., description="Whether blow-out is enabled.")
    params: BlowoutParams | SkipJsonSchema[None] = Field(
        None,
        description="Parameters for the blowout function.",
        json_schema_extra=_remove_default,
    )

    @model_validator(mode="before")
    @classmethod
    def reshape(cls, data: Any) -> Any:
        """Move any params specified as top-level keys into the 'params' value."""
        return reshape_glob(
            data=data, params_model=BlowoutParams, property_name="Blowout properties"
        )

    @field_validator("params")
    @classmethod
    def _validate_params(
        cls, v: Optional[BlowoutParams], info: ValidationInfo
    ) -> Optional[BlowoutParams]:
        if v is None and info.data.get("enable", False):
            raise ValueError(
                "If enable is true parameters for blowout must be defined."
            )
        return v


def reshape_glob(
    data: Any,
    params_model: Type[
        Union[LiquidClassTouchTipParams, BlowoutParams, DelayParams, MixParams]
    ],
    property_name: str,
) -> Any:
    """Move any params specified as top-level keys into the 'params' value.

    Also move value of 'enabled' key into 'enable' key.
    """
    # NOTE: This does not check that the input dictionary is strictly in the shape of
    # a TransferPropertiesDict. Specifically, it doesn't check that it receives the
    # 'enabled' key when params are specified as top-level keys. So something like this
    # dict will still convert successfully- {'enable': True, 'repetitions': 1, 'volume': 1}
    # To conform to TransferPropertiesDict, the enable key should be 'enabled'.
    # I am allowing this since it still can be converted to a valid LC params model.

    if isinstance(data, dict):
        if None not in (data.get("enable"), data.get("enabled")):
            raise ValueError(
                f"{property_name} should specify either 'enable' or 'enabled', not both."
            )
        if data.get("enabled") is not None:
            data["enable"] = data["enabled"]
            data.pop("enabled")

        params_list = [meta.alias for field, meta in params_model.model_fields.items()]
        list_of_presence_of_params = [param in data.keys() for param in params_list]
        if any(list_of_presence_of_params):
            if not all(list_of_presence_of_params):
                raise ValueError(
                    f"{property_name} should specify either all of the params-"
                    f"{params_list} - or none of them."
                )
            if data.get("params"):
                raise ValueError(
                    f"{property_name} should specify either all of"
                    f" {params_list} or 'params', not both."
                )
            data["params"] = params_model.model_validate(
                {param: data[param] for param in params_list}
            )
            for param in params_list:
                data.pop(param)
    return data


class Submerge(BaseLiquidClassModel):
    """Shared properties for the submerge function before aspiration or dispense."""

    startPosition: TipPosition = Field(
        ...,
        alias="start_position",
        description="Tip position before starting the submerge.",
    )
    speed: _NonNegativeNumber = Field(
        ..., description="Speed of submerging, in millimeters per second."
    )
    delay: DelayProperties = Field(..., description="Delay settings for submerge.")


class RetractAspirate(BaseLiquidClassModel):
    """Shared properties for the retract function after aspiration."""

    endPosition: TipPosition = Field(
        ..., alias="end_position", description="Tip position at the end of the retract."
    )
    speed: _NonNegativeNumber = Field(
        ..., description="Speed of retraction, in millimeters per second."
    )
    airGapByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="air_gap_by_volume",
        description="Settings for air gap keyed by target aspiration volume.",
    )
    touchTip: TouchTipProperties = Field(
        ...,
        alias="touch_tip",
        description="Touch tip settings for retract after aspirate.",
    )
    delay: DelayProperties = Field(
        ..., description="Delay settings for retract after aspirate."
    )


class RetractDispense(BaseLiquidClassModel):
    """Shared properties for the retract function after dispense."""

    endPosition: TipPosition = Field(
        ..., alias="end_position", description="Tip position at the end of the retract."
    )
    speed: _NonNegativeNumber = Field(
        ..., description="Speed of retraction, in millimeters per second."
    )
    airGapByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="air_gap_by_volume",
        description="Settings for air gap keyed by target aspiration volume.",
    )
    blowout: BlowoutProperties = Field(
        ..., description="Blowout properties for retract after dispense."
    )
    touchTip: TouchTipProperties = Field(
        ...,
        alias="touch_tip",
        description="Touch tip settings for retract after dispense.",
    )
    delay: DelayProperties = Field(
        ..., description="Delay settings for retract after dispense."
    )


class AspirateProperties(BaseLiquidClassModel):
    """Properties specific to the aspirate function."""

    submerge: Submerge = Field(..., description="Submerge settings for aspirate.")
    retract: RetractAspirate = Field(
        ..., description="Pipette retract settings after an aspirate."
    )
    aspiratePosition: TipPosition = Field(
        ..., alias="aspirate_position", description="Tip position during aspirate."
    )
    flowRateByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="flow_rate_by_volume",
        description="Settings for flow rate keyed by target aspiration volume.",
    )
    correctionByVolume: CorrectionByVolume = Field(
        ...,
        alias="correction_by_volume",
        description="Settings for volume correction keyed by by target aspiration volume,"
        " representing additional volume the plunger should move to accurately hit target volume.",
    )
    preWet: bool = Field(
        ..., alias="pre_wet", description="Whether to perform a pre-wet action."
    )
    mix: MixProperties = Field(
        ..., description="Mixing settings for before an aspirate"
    )
    delay: DelayProperties = Field(..., description="Delay settings after an aspirate")


class SingleDispenseProperties(BaseLiquidClassModel):
    """Properties specific to the single-dispense function."""

    submerge: Submerge = Field(
        ..., description="Submerge settings for single dispense."
    )
    retract: RetractDispense = Field(
        ..., description="Pipette retract settings after a single dispense."
    )
    dispensePosition: TipPosition = Field(
        ..., alias="dispense_position", description="Tip position during dispense."
    )
    flowRateByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="flow_rate_by_volume",
        description="Settings for flow rate keyed by target dispense volume.",
    )
    correctionByVolume: CorrectionByVolume = Field(
        ...,
        alias="correction_by_volume",
        description="Settings for volume correction keyed by by target dispense volume,"
        " representing additional volume the plunger should move to accurately hit target volume.",
    )
    mix: MixProperties = Field(..., description="Mixing settings for after a dispense")
    pushOutByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="push_out_by_volume",
        description="Settings for pushout keyed by target dispense volume.",
    )
    delay: DelayProperties = Field(..., description="Delay after dispense, in seconds.")


class MultiDispenseProperties(BaseLiquidClassModel):
    """Properties specific to the multi-dispense function."""

    submerge: Submerge = Field(..., description="Submerge settings for multi-dispense.")
    retract: RetractDispense = Field(
        ..., description="Pipette retract settings after a multi-dispense."
    )
    dispensePosition: TipPosition = Field(
        ..., alias="dispense_position", description="Tip position during dispense."
    )
    flowRateByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="flow_rate_by_volume",
        description="Settings for flow rate keyed by target dispense volume.",
    )
    correctionByVolume: CorrectionByVolume = Field(
        ...,
        alias="correction_by_volume",
        description="Settings for volume correction keyed by by target dispense volume,"
        " representing additional volume the plunger should move to accurately hit target volume.",
    )
    conditioningByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="conditioning_by_volume",
        description="Settings for conditioning volume keyed by target dispense volume.",
    )
    disposalByVolume: LiquidHandlingPropertyByVolume = Field(
        ...,
        alias="disposal_by_volume",
        description="Settings for disposal volume keyed by target dispense volume.",
    )
    delay: DelayProperties = Field(
        ..., description="Delay settings after each dispense"
    )


class TransferProperties(BaseLiquidClassModel):
    """Properties used during a transfer."""

    aspirate: AspirateProperties = Field(
        ..., description="Aspirate parameters for this tip type."
    )
    singleDispense: SingleDispenseProperties = Field(
        ...,
        alias="dispense",
        description="Single dispense parameters for this tip type.",
    )
    multiDispense: MultiDispenseProperties | SkipJsonSchema[None] = Field(
        None,
        alias="multi_dispense",
        description="Optional multi-dispense parameters for this tip type.",
        json_schema_extra=_remove_default,
    )


class ByTipTypeSetting(TransferProperties):
    """Settings for each kind of tip this pipette can use."""

    tiprack: str = Field(
        ...,
        description="The name of tiprack whose tip will be used when handling this specific liquid class with this pipette",
    )


class ByPipetteSetting(BaseLiquidClassModel):
    """The settings for this liquid class when used with a specific kind of pipette."""

    pipetteModel: str = Field(..., description="The pipette model this applies to.")
    byTipType: Sequence[ByTipTypeSetting] = Field(
        ..., description="Settings for each kind of tip this pipette can use"
    )


class LiquidClassSchemaV1(BaseLiquidClassModel):
    """Defines a single liquid class's properties for liquid handling functions."""

    liquidClassName: str = Field(
        ..., description="The name of the liquid (e.g., water, ethanol, serum)."
    )
    displayName: str = Field(..., description="User-readable name of the liquid class.")
    description: str = Field(
        ..., description="User-readable description of the liquid class"
    )
    schemaVersion: Literal[1] = Field(
        ..., description="Which schema version a liquid class is using"
    )
    version: int = Field(
        ..., description="Version of the specific liquid class definition"
    )
    namespace: str = Field(...)
    byPipette: Sequence[ByPipetteSetting] = Field(
        ...,
        description="Liquid class settings by each pipette compatible with this liquid class.",
    )



import os
import json
from typing import Iterator
from opentrons_shared_data import get_shared_data_root

from opentrons_shared_data.pipette.pipette_definition import (
    PipetteConfigurations,
)
from opentrons_shared_data.pipette.load_data import (
    load_definition,
    load_valid_nozzle_maps,
)
from opentrons_shared_data.pipette.pipette_load_name_conversions import (
    convert_pipette_model,
)
from opentrons_shared_data.pipette import types


def iterate_models() -> Iterator[types.PipetteModel]:
    """Get an iterator of all pipette models."""
    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }
    defn_root = get_shared_data_root() / "pipette" / "definitions" / "2" / "liquid"
    assert os.listdir(defn_root), "A path is wrong"
    for channel_dir in defn_root.iterdir():
        for model_dir in channel_dir.iterdir():
            for lc_dir in model_dir.iterdir():
                for version_file in lc_dir.iterdir():
                    version_list = version_file.stem.split("_")
                    yield types.PipetteModel(
                        f"{model_dir.stem}_{_channel_model_str[channel_dir.stem]}_v{version_list[0]}.{version_list[1]}"
                    )


def test_check_all_models_are_valid() -> None:
    """Make sure each model can be loaded."""
    for model in iterate_models():
        model_version = convert_pipette_model(model)
        try:
            loaded_model = load_definition(
                model_version.pipette_type,
                model_version.pipette_channels,
                model_version.pipette_version,
                model_version.oem_type,
            )
        except json.JSONDecodeError:
            print(
                f"Failed to parse {model_version.pipette_type} {model_version.pipette_channels} {model_version.pipette_version}"
            )
            raise

        assert isinstance(loaded_model, PipetteConfigurations)


def test_pick_up_configs_configuration_by_nozzle_map_keys() -> None:
    """Verify that for every nozzle map the pickUpTipConfigurations have a valid configuration set."""
    paths_to_validate = (
        get_shared_data_root() / "pipette" / "definitions" / "2" / "general"
    )
    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }
    assert os.listdir(paths_to_validate), "You have a path wrong"
    for channel_dir in os.listdir(paths_to_validate):
        for model_dir in os.listdir(paths_to_validate / channel_dir):
            for version_file in os.listdir(paths_to_validate / channel_dir / model_dir):
                print(version_file)
                version_list = version_file.split(".json")[0].split("_")
                built_model: types.PipetteModel = types.PipetteModel(
                    f"{model_dir}_{_channel_model_str[channel_dir]}_v{version_list[0]}.{version_list[1]}"
                )

                model_version = convert_pipette_model(built_model)
                loaded_model = load_definition(
                    model_version.pipette_type,
                    model_version.pipette_channels,
                    model_version.pipette_version,
                    model_version.oem_type,
                )
                valid_nozzle_maps = load_valid_nozzle_maps(
                    model_version.pipette_type,
                    model_version.pipette_channels,
                    model_version.pipette_version,
                    model_version.oem_type,
                )

                pipette_maps = list(
                    loaded_model.pick_up_tip_configurations.press_fit.configuration_by_nozzle_map.keys()
                )
                if loaded_model.pick_up_tip_configurations.cam_action is not None:
                    pipette_maps += list(
                        loaded_model.pick_up_tip_configurations.cam_action.configuration_by_nozzle_map.keys()
                    )
                assert pipette_maps == list(valid_nozzle_maps.maps.keys())


def test_pick_up_configs_configuration_ordered_from_smallest_to_largest() -> None:
    """Verify that ValidNozzleMaps and PickUpTipConfigurations are ordered from smallest nozzle configuration to largest, ensuring mutable config compliancy."""
    paths_to_validate = (
        get_shared_data_root() / "pipette" / "definitions" / "2" / "general"
    )
    _channel_model_str = {
        "single_channel": "single",
        "ninety_six_channel": "96",
        "eight_channel": "multi",
        "eight_channel_em": "multi_em",
    }
    assert os.listdir(paths_to_validate), "You have a path wrong"
    for channel_dir in os.listdir(paths_to_validate):
        for model_dir in os.listdir(paths_to_validate / channel_dir):
            for version_file in os.listdir(paths_to_validate / channel_dir / model_dir):
                version_list = version_file.split(".json")[0].split("_")
                built_model: types.PipetteModel = types.PipetteModel(
                    f"{model_dir}_{_channel_model_str[channel_dir]}_v{version_list[0]}.{version_list[1]}"
                )

                model_version = convert_pipette_model(built_model)
                loaded_model = load_definition(
                    model_version.pipette_type,
                    model_version.pipette_channels,
                    model_version.pipette_version,
                    model_version.oem_type,
                )
                valid_nozzle_maps = load_valid_nozzle_maps(
                    model_version.pipette_type,
                    model_version.pipette_channels,
                    model_version.pipette_version,
                    model_version.oem_type,
                )

                map_keys = list(valid_nozzle_maps.maps.keys())
                for key in map_keys:
                    for i in range(map_keys.index(key)):
                        assert len(valid_nozzle_maps.maps[map_keys[i]]) <= len(
                            valid_nozzle_maps.maps[key]
                        )

                pipette_maps = list(
                    loaded_model.pick_up_tip_configurations.press_fit.configuration_by_nozzle_map.keys()
                )
                for key in pipette_maps:
                    for i in range(pipette_maps.index(key)):
                        assert len(valid_nozzle_maps.maps[pipette_maps[i]]) <= len(
                            valid_nozzle_maps.maps[key]
                        )


def test_serializer() -> None:
    """Verify that the serializer works as expected."""

    loaded_model = load_definition(
        types.PipetteModelType.p1000,
        types.PipetteChannelType.NINETY_SIX_CHANNEL,
        types.PipetteVersionType(3, 3),
        types.PipetteOEMType.OT,
    )
    quirk_0 = types.Quirks.pickupTipShake
    quirk_1 = types.Quirks.dropTipShake
    loaded_model.quirks = [quirk_0, quirk_1]

    assert loaded_model.pipette_type == types.PipetteModelType.p1000
    assert loaded_model.display_category == types.PipetteGenerationType.FLEX
    assert loaded_model.channels == types.PipetteChannelType.NINETY_SIX_CHANNEL

    model_dict = loaded_model.model_dump()
    # each field should be the value of the enum class
    assert (
        isinstance(model_dict["pipette_type"], str)
        and model_dict["pipette_type"] == loaded_model.pipette_type.value
    )
    assert (
        isinstance(model_dict["display_category"], str)
        and model_dict["display_category"] == loaded_model.display_category.value
    )
    assert (
        isinstance(model_dict["channels"], int)
        and model_dict["channels"] == loaded_model.channels.value
    )

    assert len(model_dict["quirks"]) == 2
    dict_quirk_0 = model_dict["quirks"][0]
    dict_quirk_1 = model_dict["quirks"][1]
    assert isinstance(dict_quirk_0, str) and dict_quirk_0 == quirk_0.value
    assert isinstance(dict_quirk_1, str) and dict_quirk_1 == quirk_1.value


# TODO: (AA, 4/9/2024) we should add a test to validate the dumped json to make
# sure we can re-load it as the BaseModel class. But we haven't added serializer
# for other enums yet, such as LiquidClass, and since we haven't been
# creating the definition files using model_dump/model_dump_json, it is okay to
# skip this for now.



import pytest

from opentrons_shared_data.deck import list_names


@pytest.mark.parametrize("version", [3, 4])
def test_list_names(version: int) -> None:
    """Make sure `list_names()` returns something.

    Just a basic test to make sure it's not looking in a nonexistent directory or something.
    """
    assert len(list_names(version)) > 0



import json
from typing import Any, Dict, Union
from .types import MutableConfig, QuirkConfig
from .model_constants import VALID_QUIRKS


DecoderType = Union[Dict[str, Any], Dict[str, QuirkConfig], MutableConfig]


class MutableConfigurationEncoder(json.JSONEncoder):
    def default(self, obj: object) -> Any:
        if isinstance(obj, MutableConfig) or isinstance(obj, QuirkConfig):
            return obj.dict_for_encode()
        return json.JSONEncoder.default(self, obj)


class MutableConfigurationDecoder(json.JSONDecoder):
    def __init__(self) -> None:
        super().__init__(
            object_hook=self.dict_to_obj,
        )

    def dict_to_obj(self, d: object) -> DecoderType:
        if isinstance(d, dict):
            converted = dict(d)
            if converted.get("value"):
                return self._decode_mutable_config("", converted)
            elif all(isinstance(v, bool) for v in converted.values()):
                return {
                    q: QuirkConfig.validate_and_build(q, b)
                    for q, b in converted.items()
                    if q in VALID_QUIRKS
                }
        return d  # type: ignore

    def _decode_mutable_config(self, k: str, obj: Dict[str, Any]) -> MutableConfig:
        return MutableConfig.build(**obj, name=k)



"""Protocol file reading interfaces."""
from . import protocol_schema_v6, protocol_schema_v7, protocol_schema_v8
from .protocol_schema_v6 import ProtocolSchemaV6
from .protocol_schema_v7 import ProtocolSchemaV7
from .protocol_schema_v8 import ProtocolSchemaV8, CommandSchemaId
from .shared_models import (
    Liquid,
    Labware,
    CommandAnnotation,
    Location,
    ProfileStep,
    WellLocation,
    OffsetVector,
    Dimensions,
    GroupMetadata,
    Shape,
    WellDefinition,
    Metadata,
    Module,
    Pipette,
    Robot,
    DesignerApplication,
)

__all__ = [
    "ProtocolSchemaV6",
    "protocol_schema_v6",
    "ProtocolSchemaV7",
    "protocol_schema_v7",
    "ProtocolSchemaV8",
    "protocol_schema_v8",
    "Liquid",
    "Labware",
    "CommandAnnotation",
    "Location",
    "ProfileStep",
    "WellLocation",
    "OffsetVector",
    "Dimensions",
    "GroupMetadata",
    "Shape",
    "WellDefinition",
    "Metadata",
    "Module",
    "Pipette",
    "Robot",
    "DesignerApplication",
    "CommandSchemaId",
]



"""Test that our bindings can validate and parse our standard labware definitions."""

from typing import Literal

import pytest
import typeguard
import pydantic

from opentrons_shared_data.labware import load_definition
from opentrons_shared_data.labware.labware_definition import (
    labware_definition_type_adapter as pydantic_labware_definition_type_adapter,
    LabwareDefinition2 as PydanticLabwareDefinition2,
    LabwareDefinition3 as PydanticLabwareDefinition3,
)
from opentrons_shared_data.labware.types import (
    LabwareDefinition as TypedDictLabwareDefinition,
    LabwareDefinition2 as TypedDictLabwareDefinition2,
    LabwareDefinition3 as TypedDictLabwareDefinition3,
)


from . import get_ot_defs


@pytest.mark.parametrize("loadname,version", get_ot_defs(schema=2))
def test_schema_2_types(loadname: str, version: int) -> None:
    """Test parsing and validating into the types that represent schema 2."""
    defdict = load_definition(loadname, version, schema=2)

    typeguard.check_type(defdict, TypedDictLabwareDefinition2)
    PydanticLabwareDefinition2.model_validate(defdict)


@pytest.mark.parametrize("loadname,version", get_ot_defs(schema=3))
def test_schema_3_types(loadname: str, version: int) -> None:
    """Test parsing and validating into the types that represent schema 3."""
    defdict = load_definition(loadname, version, schema=3)

    typeguard.check_type(defdict, TypedDictLabwareDefinition3)
    PydanticLabwareDefinition3.model_validate(defdict)


@pytest.mark.parametrize(
    "loadname,version,schema_version",
    [(loadname, version, 2) for loadname, version in get_ot_defs(schema=2)]
    + [(loadname, version, 3) for loadname, version in get_ot_defs(schema=3)],
)
def test_all_schema_union_types(
    loadname: str, version: int, schema_version: Literal[2, 3]
) -> None:
    """Test parsing and validating into the types that represent a union of all schemas."""
    defdict = load_definition(
        loadname=loadname,
        version=version,
        schema=schema_version,
    )

    typeguard.check_type(defdict, TypedDictLabwareDefinition)

    pydantic_result = pydantic_labware_definition_type_adapter.validate_python(defdict)
    expected_result_type = (
        PydanticLabwareDefinition2
        if schema_version == 2
        else PydanticLabwareDefinition3
    )
    assert isinstance(pydantic_result, expected_result_type)


def test_loadname_regex_applied() -> None:
    defdict = load_definition(*get_ot_defs(schema=2)[0])
    defdict["parameters"]["loadName"] = "ALSJHDAKJLA"
    with pytest.raises(pydantic.ValidationError):
        PydanticLabwareDefinition2.model_validate(defdict)


def test_namespace_regex_applied() -> None:
    defdict = load_definition(*get_ot_defs(schema=2)[0])
    defdict["namespace"] = "ALSJHDAKJLA"
    with pytest.raises(pydantic.ValidationError):
        PydanticLabwareDefinition2.model_validate(defdict)



import enum
from dataclasses import dataclass

from pydantic import BaseModel
from typing_extensions import Literal, TypedDict
from typing import Dict, List, Mapping, NewType, Union, Tuple, cast


# TODO(mc, 2022-06-16): remove type alias when able
# and when certain removal will not break any pickling
from ..labware.types import LabwareUri as LabwareUri

"""Pipette Definition V2 Types"""

# Needed for Int Comparison. Keeping it next to
# the Literal type for ease of readability
PipetteModelMajorVersion = [1, 2, 3]
PipetteModelMinorVersion = [0, 1, 2, 3, 4, 5, 6, 7]

# TODO Literals are only good for writing down
# exact values. Is there a better typing mechanism
# so we don't need to keep track of versions in two
# different places?
PipetteModelMajorVersionType = Literal[1, 2, 3]
PipetteModelMinorVersionType = Literal[0, 1, 2, 3, 4, 5, 6, 7]


class LiquidClasses(enum.Enum):
    default = enum.auto()
    lowVolumeDefault = enum.auto()


class Vector(BaseModel):
    x: float
    y: float
    z: float


LIQUID_PROBE_START_OFFSET_FROM_WELL_TOP = Vector(x=0, y=0, z=2)
"""The (x, y, z) offset from well top to use as start location for probing liquid in that well."""


class PipetteTipType(enum.Enum):
    t10 = 10
    t20 = 20
    t50 = 50
    t200 = 200
    t300 = 300
    t1000 = 1000

    @classmethod
    def check_and_return_type(
        cls, working_volume: int, maximum_volume: int
    ) -> "PipetteTipType":
        try:
            return cls(working_volume)
        except ValueError:
            return cls(maximum_volume)


class PipetteChannelType(int, enum.Enum):
    SINGLE_CHANNEL = 1
    EIGHT_CHANNEL = 8
    NINETY_SIX_CHANNEL = 96

    def __str__(self) -> str:
        if self.value == 96:
            return "96"
        elif self.value == 8:
            return "multi"
        else:
            return "single"


class PipetteModelType(enum.Enum):
    p10 = "p10"
    p20 = "p20"
    p50 = "p50"
    p200 = "p200"
    p300 = "p300"
    p1000 = "p1000"


class PipetteGenerationType(enum.Enum):
    GEN1 = "GEN1"
    GEN2 = "GEN2"
    FLEX = "FLEX"


@dataclass(frozen=True)
class PipetteVersionType:
    major: PipetteModelMajorVersionType
    minor: PipetteModelMinorVersionType

    @classmethod
    def convert_from_float(cls, version: float) -> "PipetteVersionType":
        major = cast(PipetteModelMajorVersionType, int(version // 1))
        minor = cast(PipetteModelMinorVersionType, int(round((version % 1), 2) * 10))
        return cls(major=major, minor=minor)

    def __str__(self) -> str:
        if self.major == 1 and self.minor == 0:
            # Maintain the format of V1 pipettes that
            # do not contain a minor version at all.
            return f"{self.major}"
        else:
            return f"{self.major}.{self.minor}"

    @property
    def as_tuple(
        self,
    ) -> Tuple[PipetteModelMajorVersionType, PipetteModelMinorVersionType]:
        return (self.major, self.minor)


"""Mutable Configs Types"""


class Quirks(enum.Enum):
    pickupTipShake = "pickupTipShake"
    dropTipShake = "dropTipShake"
    doubleDropTip = "doubleDropTip"
    needsUnstick = "needsUnstick"
    highSpeed = "highSpeed"


class PipetteOEMType(enum.Enum):
    OT = "ot"  # opentrons type
    EM = "em"  # Emulsifying Pipette

    @classmethod
    def get_oem_from_quirks(cls, quirks: List[Quirks]) -> "PipetteOEMType":
        """Return an oem type if true based on the quirks."""
        return cls.EM if Quirks.highSpeed in quirks else cls.OT

    @classmethod
    def get_oem_from_model_str(cls, model_str: str) -> "PipetteOEMType":
        """Return an oem type if true based on the model string."""
        return cls.EM if "multi_em" in model_str else cls.OT


class AvailableUnits(enum.Enum):
    mm = "mm"
    amps = "amps"
    speed = "mm/s"
    presses = "presses"


@dataclass
class RobotMountConfigs:
    stepsPerMM: float
    homePosition: float
    travelDistance: float


@dataclass
class MutableConfig:
    value: Union[int, float]
    default: Union[int, float]
    units: AvailableUnits
    type: str
    min: float
    max: float
    name: str

    @classmethod
    def build(
        cls,
        value: Union[int, float],
        default: Union[int, float],
        units: str,
        type: str,
        min: float,
        max: float,
        name: str,
    ) -> "MutableConfig":
        if units == "mm/sec":
            units = "mm/s"
        return cls(
            value=value,
            default=default,
            units=AvailableUnits(units),
            type=type,
            min=min,
            max=max,
            name=name,
        )

    def validate_and_add(self, new_value: Union[int, float]) -> None:
        if new_value < self.min or new_value > self.max:
            raise ValueError(f"{self.name} out of range with {new_value}")
        self.value = new_value

    def dict_for_encode(self) -> Dict[str, Union[int, float, str]]:
        return {
            "value": self.value,
            "default": self.default,
            "units": self.units.value,
            "type": self.type,
            "min": self.min,
            "max": self.max,
        }


@dataclass
class QuirkConfig:
    value: bool
    name: Quirks

    @classmethod
    def validate_and_build(cls, name: str, value: bool) -> "QuirkConfig":
        quirk_name = Quirks(name)
        cls.validate(name, value)
        return cls(value=value, name=quirk_name)

    @classmethod
    def validate(cls, name: str, value: bool) -> None:
        if not isinstance(value, bool):
            raise ValueError(f"{value} is invalid for {name}")

    def dict_for_encode(self) -> bool:
        return self.value


TypeOverrides = Mapping[str, Union[float, bool, None]]

OverrideType = Dict[str, Union[Dict[str, QuirkConfig], MutableConfig, str]]

PipetteName = Literal[
    "p10_single",
    "p10_multi",
    "p20_single_gen2",
    "p20_multi_gen2",
    "p50_single",
    "p50_multi",
    "p50_single_flex",
    "p50_multi_flex",
    "p300_single",
    "p300_multi",
    "p300_single_gen2",
    "p300_multi_gen2",
    "p1000_single",
    "p1000_single_gen2",
    "p1000_single_flex",
    "p1000_multi_flex",
    "p1000_multi_em_flex",
    "p1000_96",
    "p200_96",
]


class PipetteNameType(str, enum.Enum):
    """Pipette load name values."""

    value: PipetteName

    P10_SINGLE = "p10_single"
    P10_MULTI = "p10_multi"
    P20_SINGLE_GEN2 = "p20_single_gen2"
    P20_MULTI_GEN2 = "p20_multi_gen2"
    P50_SINGLE = "p50_single"
    P50_MULTI = "p50_multi"
    P50_SINGLE_FLEX = "p50_single_flex"
    P50_MULTI_FLEX = "p50_multi_flex"
    P300_SINGLE = "p300_single"
    P300_MULTI = "p300_multi"
    P300_SINGLE_GEN2 = "p300_single_gen2"
    P300_MULTI_GEN2 = "p300_multi_gen2"
    P1000_SINGLE = "p1000_single"
    P1000_SINGLE_GEN2 = "p1000_single_gen2"
    P1000_SINGLE_FLEX = "p1000_single_flex"
    P1000_MULTI_FLEX = "p1000_multi_flex"
    P1000_MULTI_EM = "p1000_multi_em_flex"
    P1000_96 = "p1000_96"
    P200_96 = "p200_96"


# Mapping of public Python Protocol API pipette load names
# to names used by the internal Opentrons system
# TODO (spp, 2025-01-02): make the API load name a part of the pipette's definition
PIPETTE_API_NAMES_MAP = {
    "p10_single": PipetteNameType.P10_SINGLE,
    "p10_multi": PipetteNameType.P10_MULTI,
    "p20_single_gen2": PipetteNameType.P20_SINGLE_GEN2,
    "p20_multi_gen2": PipetteNameType.P20_MULTI_GEN2,
    "p50_single": PipetteNameType.P50_SINGLE,
    "p50_multi": PipetteNameType.P50_MULTI,
    "p300_single": PipetteNameType.P300_SINGLE,
    "p300_multi": PipetteNameType.P300_MULTI,
    "p300_single_gen2": PipetteNameType.P300_SINGLE_GEN2,
    "p300_multi_gen2": PipetteNameType.P300_MULTI_GEN2,
    "p1000_single": PipetteNameType.P1000_SINGLE,
    "p1000_single_gen2": PipetteNameType.P1000_SINGLE_GEN2,
    "flex_1channel_50": PipetteNameType.P50_SINGLE_FLEX,
    "flex_8channel_50": PipetteNameType.P50_MULTI_FLEX,
    "flex_1channel_1000": PipetteNameType.P1000_SINGLE_FLEX,
    "flex_8channel_1000": PipetteNameType.P1000_MULTI_FLEX,
    "flex_8channel_1000_em": PipetteNameType.P1000_MULTI_EM,
    "flex_96channel_1000": PipetteNameType.P1000_96,
    "flex_96channel_200": PipetteNameType.P200_96,
}


# Generic NewType for models because we get new ones frequently and theres
# a huge number of them
PipetteModel = NewType("PipetteModel", str)

DisplayCategory = Literal["GEN1", "GEN2", "FLEX"]

# todo(mm, 2022-03-18):
# The JSON schema defines this as any string, not as an enum of string literals.
# Check if it's safe to simplify this to just str.
ConfigUnit = Literal[
    "mm",
    "amps",
    "mm/sec",
    "mm/s",  # todo(mm, 2022-03-18): Standardize specs to mm/sec or mm/s.
    "presses",
]

Quirk = NewType("Quirk", str)

ChannelCount = Literal[1, 8, 96]

UlPerMmAction = Literal["aspirate", "dispense", "blowout"]


class PipetteConfigElement(TypedDict):
    value: float
    min: float
    max: float


class PipetteConfigElementWithPerApiLevelValue(TypedDict):
    value: float
    min: float
    max: float
    valuesByApiLevel: Dict[str, float]


# TypedDicts can't be generic sadly
class PipetteCustomizableConfigElementFloat(TypedDict):
    value: float
    min: float
    max: float
    units: ConfigUnit
    type: Literal["float"]


class PipetteCustomizableConfigElementInt(TypedDict):
    value: int
    min: int
    max: int
    units: ConfigUnit
    type: Literal["int"]


PipetteCustomizableConfigElement = Union[
    PipetteCustomizableConfigElementFloat, PipetteCustomizableConfigElementInt
]

SmoothieConfigs = TypedDict(
    "SmoothieConfigs",
    {"stepsPerMM": float, "homePosition": float, "travelDistance": float},
)


class PipetteNameSpec(TypedDict):
    displayName: str
    displayCategory: DisplayCategory
    minVolume: Union[float, int]
    maxVolume: Union[float, int]
    channels: ChannelCount
    defaultAspirateFlowRate: PipetteConfigElementWithPerApiLevelValue
    defaultDispenseFlowRate: PipetteConfigElementWithPerApiLevelValue
    defaultBlowOutFlowRate: PipetteConfigElementWithPerApiLevelValue
    smoothieConfigs: SmoothieConfigs
    defaultTipracks: List[LabwareUri]


PipetteNameSpecs = Dict[PipetteName, PipetteNameSpec]

UlPerMm = Dict[UlPerMmAction, List[List[float]]]


class PipetteModelSpec(TypedDict, total=False):
    name: PipetteName
    top: PipetteCustomizableConfigElementFloat
    bottom: PipetteCustomizableConfigElementFloat
    blowout: PipetteCustomizableConfigElementFloat
    dropTip: PipetteCustomizableConfigElementFloat
    pickUpCurrent: PipetteCustomizableConfigElementFloat
    pickUpDistance: PipetteCustomizableConfigElementFloat
    pickUpIncrement: PipetteCustomizableConfigElementFloat
    pickUpPresses: PipetteCustomizableConfigElementInt
    pickUpSpeed: PipetteCustomizableConfigElementFloat
    plungerCurrent: PipetteCustomizableConfigElementFloat
    dropTipCurrent: PipetteCustomizableConfigElementFloat
    dropTipSpeed: PipetteCustomizableConfigElementFloat
    modelOffset: List[float]
    nozzleOffset: List[float]
    ulPerMm: List[UlPerMm]
    tipOverlap: Dict[str, float]
    tipLength: PipetteCustomizableConfigElementFloat
    quirks: List[Quirk]
    # these keys are not present in some pipette definitions
    backCompatNames: List[PipetteName]
    idleCurrent: float
    returnTipHeight: float


class PipetteFusedSpec(PipetteNameSpec, PipetteModelSpec, total=False):
    pass


class PipetteModelSpecs(TypedDict):
    config: Dict[PipetteModel, PipetteModelSpec]
    mutableConfigs: List[str]
    validQuirks: List[str]



import json
import logging
import os
from pathlib import Path
from typing import Dict, Any, cast, Union, Generator

import pytest

from opentrons_shared_data.pipette import (
    mutable_configurations,
    types,
    pipette_definition,
    pipette_load_name_conversions as pip_conversions,
    load_data,
)


TEST_SERIAL_NUMBER = "P50MV1520200304"
TestOverrideType = Dict[str, Union[float, int, bool]]


@pytest.fixture
def TMPFILE_DATA() -> Dict[str, Any]:
    return {
        "dropTipShake": True,
        "doubleDropTip": True,
        "model": "p50_multi_v1.5",
        "quirks": {"pickUpPresses": True, "dropTipShake": True, "doubleDropTip": True},
        "pickUpSpeed": {
            "value": 5.0,
            "min": 1,
            "max": 100,
            "units": "mm/s",
            "type": "float",
            "default": 30,
        },
    }


@pytest.fixture
def override_configuration_path(tmp_path: Path) -> Generator[Path, None, None]:
    os.environ["OT_API_CONFIG_DIR"] = str(tmp_path)

    tmp_path.mkdir(parents=True, exist_ok=True)
    with_pip_path = tmp_path / Path("pipettes")
    with_pip_path.mkdir(parents=True, exist_ok=True)
    yield with_pip_path

    del os.environ["OT_API_CONFIG_DIR"]


@pytest.fixture
def overrides_fixture(
    override_configuration_path: Path, TMPFILE_DATA: Dict[str, Any]
) -> types.MutableConfig:
    with open(override_configuration_path / f"{TEST_SERIAL_NUMBER}.json", "w") as f:
        json.dump(TMPFILE_DATA, f)
    return types.MutableConfig.build(**TMPFILE_DATA["pickUpSpeed"], name="pickUpSpeed")


def test_load_old_overrides_regression(
    TMPFILE_DATA: Dict[str, Any], override_configuration_path: Path
) -> None:
    TMPFILE_DATA["pickUpCurrent"] = {
        "value": 0.15,
        "min": 0.08,
        "max": 0.2,
        "units": "amps",
        "type": "float",
        "default": 0.1,
    }
    with open(override_configuration_path / "P20SV222021040709.json", "w") as f:
        json.dump(TMPFILE_DATA, f)
    configs = mutable_configurations.load_with_mutable_configurations(
        pipette_definition.PipetteModelVersionType(
            pipette_type=types.PipetteModelType.p20,
            pipette_channels=types.PipetteChannelType.SINGLE_CHANNEL,
            pipette_version=types.PipetteVersionType(2, 2),
            oem_type=types.PipetteOEMType.OT,
        ),
        override_configuration_path,
        "P20SV222021040709",
    )
    assert (
        configs.pick_up_tip_configurations.press_fit.configuration_by_nozzle_map[
            list(
                configs.pick_up_tip_configurations.press_fit.configuration_by_nozzle_map.keys()
            )[0]
        ]["default"].current
        == 0.15
    )


def test_list_mutable_configs_unknown_pipette_id(
    override_configuration_path: Path,
) -> None:
    """Test unknown pipette id mutable configs.

    Test that a user receives a list of all possible mutable configurations
    with the default value equal to the regular value.
    """

    found_configurations = mutable_configurations.list_mutable_configs(
        TEST_SERIAL_NUMBER, override_configuration_path
    )
    for c in found_configurations:
        if isinstance(c, str):
            # model string, ignore
            continue
        if isinstance(c, types.QuirkConfig):
            assert isinstance(c.value, bool)
        else:
            assert c.default == c.value


def test_list_mutable_configs_known_pipette_id(
    overrides_fixture: types.MutableConfig, override_configuration_path: Path
) -> None:
    """Test known pipette id mutable configs.

    Test that a user receives a list of all possible mutable configurations
    with the expected overrides also listed.
    """
    found_configurations = mutable_configurations.list_mutable_configs(
        TEST_SERIAL_NUMBER, override_configuration_path
    )

    for c in found_configurations:
        if isinstance(c, str):
            # model string, ignore
            continue
        if overrides_fixture.name == c.name:
            assert c.value == overrides_fixture.value
        elif isinstance(c, types.QuirkConfig):
            assert isinstance(c.value, bool)
        else:
            assert c.default == c.value


@pytest.mark.parametrize(
    argnames=["overrides_dict", "saved_dict"],
    argvalues=[
        [
            {"pickUpCurrent": 0.5, "dropTipSpeed": 5, "dropTipShake": False},
            {
                "quirks": {"dropTipShake": False},
                "pickUpCurrent": {
                    "value": 0.5,
                    "default": 0.8,
                    "units": "amps",
                    "type": "float",
                    "min": 0.1,
                    "max": 2.0,
                },
                "model": "p50_multi_v1.5",
                "dropTipSpeed": {
                    "value": 5,
                    "default": 5.0,
                    "units": "mm/s",
                    "type": "float",
                    "min": 0.01,
                    "max": 30,
                },
            },
        ]
    ],
)
def test_save_new_overrides_new_file(
    override_configuration_path: Path,
    overrides_dict: TestOverrideType,
    saved_dict: Dict[str, Any],
) -> None:
    mutable_configurations.save_overrides(
        TEST_SERIAL_NUMBER, overrides_dict, override_configuration_path
    )
    with open(override_configuration_path / f"{TEST_SERIAL_NUMBER}.json") as f:
        new_file = json.load(f)
    assert saved_dict == new_file


@pytest.mark.parametrize(
    argnames=["overrides_dict"],
    argvalues=[
        [{"pickUpCurrent": 1, "pickUpSpeed": 10, "dropTipShake": False}],
        [{"pickUpCurrent": 2}],
    ],
)
def test_save_new_overrides_update_file(
    override_configuration_path: Path,
    overrides_fixture: types.MutableConfig,
    overrides_dict: TestOverrideType,
    TMPFILE_DATA: Dict[str, Any],
) -> None:
    mutable_configurations.save_overrides(
        TEST_SERIAL_NUMBER, overrides_dict, override_configuration_path
    )
    with open(override_configuration_path / f"{TEST_SERIAL_NUMBER}.json") as f:
        new_file = json.load(f)

    for k, v in overrides_dict.items():
        if isinstance(v, bool):
            TMPFILE_DATA["quirks"][k] = v
        elif TMPFILE_DATA.get(k):
            TMPFILE_DATA[k]["value"] = v

    TMPFILE_DATA["pickUpCurrent"] = {
        "default": 0.8,
        "max": 2.0,
        "min": 0.1,
        "type": "float",
        "units": "amps",
        "value": overrides_dict["pickUpCurrent"],
    }

    del TMPFILE_DATA["quirks"]["pickUpPresses"]
    assert TMPFILE_DATA == new_file


@pytest.mark.parametrize(
    argnames=["overrides_dict"],
    argvalues=[
        [{"pickUpCurrent": 1231.213, "dropTipSpeed": 121, "dropTipShake": False}],
        [{"quirk123": True}],
    ],
)
def test_save_invalid_overrides(
    overrides_fixture: types.MutableConfig,
    override_configuration_path: Path,
    overrides_dict: TestOverrideType,
    TMPFILE_DATA: Dict[str, Any],
) -> None:
    with pytest.raises(ValueError):
        mutable_configurations.save_overrides(
            TEST_SERIAL_NUMBER, overrides_dict, override_configuration_path
        )
    with open(override_configuration_path / f"{TEST_SERIAL_NUMBER}.json") as f:
        new_file = json.load(f)
    assert TMPFILE_DATA == new_file


@pytest.mark.parametrize(
    argnames=["pipette_model", "serial_number"],
    argvalues=[
        [
            pip_conversions.convert_pipette_model(
                cast(types.PipetteModel, "p1000_96_v3.3")
            ),
            "P1KHV3320230629",
        ],
        [
            pip_conversions.convert_pipette_model(
                cast(types.PipetteModel, "p50_multi_v1.5")
            ),
            TEST_SERIAL_NUMBER,
        ],
    ],
)
def test_load_with_overrides(
    overrides_fixture: types.MutableConfig,
    pipette_model: pipette_definition.PipetteModelVersionType,
    serial_number: str,
    override_configuration_path: Path,
) -> None:
    """Test that you can load configurations both with pre-existing overrides and non-pre-existing overrides."""
    updated_configurations = mutable_configurations.load_with_mutable_configurations(
        pipette_model, override_configuration_path, serial_number
    )

    loaded_base_configurations = load_data.load_definition(
        pipette_model.pipette_type,
        pipette_model.pipette_channels,
        pipette_model.pipette_version,
        pipette_model.oem_type,
    )

    if serial_number == TEST_SERIAL_NUMBER:
        dict_loaded_configs = loaded_base_configurations.model_dump(by_alias=True)
        for map_key in dict_loaded_configs["pickUpTipConfigurations"]["pressFit"][
            "configurationsByNozzleMap"
        ]:
            for tip_key in dict_loaded_configs["pickUpTipConfigurations"]["pressFit"][
                "configurationsByNozzleMap"
            ][map_key]:
                dict_loaded_configs["pickUpTipConfigurations"]["pressFit"][
                    "configurationsByNozzleMap"
                ][map_key][tip_key]["speed"] = 5.0

        updated_configurations_dict = updated_configurations.model_dump(by_alias=True)
        assert set(dict_loaded_configs.pop("quirks")) == set(
            updated_configurations_dict.pop("quirks")
        )
        assert updated_configurations_dict == dict_loaded_configs
    else:
        assert updated_configurations == loaded_base_configurations


def test_build_mutable_config_using_old_units() -> None:
    """Test that MutableConfigs can build with old units."""
    old_units_config = {
        "value": 5,
        "default": 5.0,
        "units": "mm/s",
        "type": "float",
        "min": 0.01,
        "max": 30,
    }
    assert (
        types.MutableConfig.build(**old_units_config, name="dropTipSpeed") is not None  # type: ignore
    )


@pytest.mark.parametrize(
    ("filename", "type", "channels", "version", "file_contents"),
    # From https://opentrons.atlassian.net/browse/RQA-3676.
    # These could probably be pared down.
    [
        (
            "P20MV202020121412.json",
            types.PipetteModelType.p20,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p20_multi_v2.0"}',
        ),
        (
            "P3HSV1318071638.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(1, 3),
            '{"dropTipShake": true, "model": "p300_single_v1.3", "quirks": {"dropTipShake": true}, "top": {"value": 30.0, "default": 19.5, "units": "mm", "type": "float", "min": -20, "max": 30}, "pickUpPresses": {"value": 3.0, "default": 3, "units": "presses", "type": "int", "min": 0, "max": 15}}',
        ),
        (
            "P3HMV212021040004.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 1),
            '{"needsUnstick": true, "model": "p300_multi_v2.1"}',
        ),
        (
            "P20SV202020032604.json",
            types.PipetteModelType.p20,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p20_single_v2.0"}',
        ),
        (
            "P1KSV202019072441.json",
            types.PipetteModelType.p1000,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"pickupTipShake": true, "model": "p1000_single_v2.0", "quirks": {"pickupTipShake": true}}',
        ),
        (
            "P3HMV202021011105.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"needsUnstick": true, "model": "p300_multi_v2.0"}',
        ),
        (
            "P20SV202019072527.json",
            types.PipetteModelType.p20,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p20_single_v2.0"}',
        ),
        (
            "P3HSV202021042602.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p300_single_v2.0"}',
        ),
        (
            "P20SV222021030914.json",
            types.PipetteModelType.p20,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 2),
            '{"model": "p20_single_v2.2", "quirks": {}, "pickUpPresses": {"value": 3.0, "default": 1, "units": "presses", "type": "int", "min": 0, "max": 15}}',
        ),
        (
            "P3HMV202020040801.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"needsUnstick": true, "model": "p300_multi_v2.0", "quirks": {"needsUnstick": true}, "top": {"value": 20.0, "default": 19.5, "units": "mm", "type": "float", "min": -20, "max": 30}, "bottom": {"value": -14.0, "default": -14.5, "units": "mm", "type": "float", "min": -20, "max": 30}, "blowout": {"value": -18.5, "default": -19.0, "units": "mm", "type": "float", "min": -20, "max": 30}, "dropTip": {"value": -20.0, "default": -33.4, "units": "mm", "type": "float", "min": -20, "max": 30}, "pickUpCurrent": {"value": 0.9, "default": 0.8, "units": "amps", "type": "float", "min": 0.1, "max": 2.0}, "pickUpDistance": {"value": 10.0, "default": 11.0, "units": "mm", "type": "float", "min": 0, "max": 10}, "pickUpIncrement": {"value": 0.1, "default": 0.0, "units": "mm", "type": "int", "min": 0, "max": 10}, "pickUpPresses": {"value": 4.0, "default": 1, "units": "presses", "type": "int", "min": 0, "max": 15}, "pickUpSpeed": {"value": 11.0, "default": 10.0, "units": "mm/s", "type": "float", "min": 0.01, "max": 30}, "plungerCurrent": {"value": 1.1, "default": 1.0, "units": "amps", "type": "float", "min": 0.1, "max": 2.0}, "dropTipCurrent": {"value": 1.22, "default": 1.25, "units": "amps", "type": "float", "min": 0.1, "max": 2.0}, "dropTipSpeed": {"value": 8.0, "default": 7.5, "units": "mm/s", "type": "float", "min": 0.01, "max": 30}, "tipLength": {"value": 52.0, "default": 51.0, "units": "mm", "type": "float", "min": 0, "max": 100}}',
        ),
        (
            "P3HMV212021040002.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 1),
            '{"needsUnstick": true, "model": "p300_multi_v2.1"}',
        ),
        (
            "P3HMV1318072625.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(1, 3),
            '{"dropTipShake": true, "model": "p300_multi_v1.3", "quirks": {"dropTipShake": true}, "pickUpPresses": {"value": 4.0, "default": 3, "units": "presses", "type": "int", "min": 0, "max": 15}}',
        ),
        (
            "P50MV1519091757.json",
            types.PipetteModelType.p50,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(1, 5),
            '{"dropTipShake": true, "doubleDropTip": true, "model": "p50_multi_v1.5", "quirks": {"doubleDropTip": true, "dropTipShake": true}}',
        ),
        (
            "P3HSV202019072224.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p300_single_v2.0", "quirks": {}}',
        ),
        (
            "P20MV202019112708.json",
            types.PipetteModelType.p20,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"model": "p20_multi_v2.0", "quirks": {}}',
        ),
        (
            "P3HSV202021031503.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 1),
            '{"model": "p300_single_v2.1"}',
        ),
        (
            "P1KSV202020060206.json",
            types.PipetteModelType.p1000,
            types.PipetteChannelType.SINGLE_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"pickupTipShake": true, "model": "p1000_single_v2.0"}',
        ),
        (
            "P3HMV202021010906.json",
            types.PipetteModelType.p300,
            types.PipetteChannelType.EIGHT_CHANNEL,
            types.PipetteVersionType(2, 0),
            '{"needsUnstick": true, "model": "p300_multi_v2.0"}',
        ),
    ],
)
def test_loading_does_not_log_warnings(
    filename: str,
    type: types.PipetteModelType,
    channels: types.PipetteChannelType,
    version: types.PipetteVersionType,
    file_contents: str,
    caplog: pytest.LogCaptureFixture,
    override_configuration_path: Path,
) -> None:
    """Make sure load_with_mutable_configurations() doesn't log any exceptions.

    load_with_mutable_configurations() suppresses and logs internal exceptions to
    protect its caller, but those are still bugs, and we still want tests to catch them.
    """
    model = pipette_definition.PipetteModelVersionType(
        type, channels, version, types.PipetteOEMType.OT
    )
    (override_configuration_path / filename).write_text(file_contents)
    with caplog.at_level(logging.WARNING):
        mutable_configurations.load_with_mutable_configurations(
            model, override_configuration_path, Path(filename).stem
        )
    assert len(caplog.records) == 0



"""
opentrons_shared_data.protocol.types: types for json protocols
"""

from typing import Any, Dict, List, Optional, Union
from enum import Enum

from typing_extensions import TypedDict, Literal
from ..pipette.types import PipetteName
from ..labware.types import LabwareDefinition2
from ..module.types import ModuleModel

SlotSpan = Literal["span7_8_10_11"]


class BlowoutLocation(Enum):
    SOURCE = "source well"
    DEST = "destination well"
    TRASH = "trash"


LiquidHandlingCommand = Union[
    Literal["transfer"], Literal["consolidate"], Literal["distribute"]
]


class LabwareRequirement(TypedDict):
    slot: str
    definitionId: str
    displayName: str


Metadata = TypedDict(
    "Metadata",
    {
        "protocolName": str,
        "protocol-name": str,
        "author": str,
        "description": Optional[str],
        "created": int,
        "lastModified": Optional[int],
        "last-modified": Optional[int],
        "category": Optional[str],
        "subcategory": Optional[str],
        "tags": List[str],
    },
    total=False,
)

AspirateCommandId = Literal["aspirate"]
DispenseCommandId = Literal["dispense"]
AirGapCommandId = Literal["airGap"]
BlowoutCommandId = Literal["blowout"]
TouchTipCommandId = Literal["touchTip"]
PickUpTipCommandId = Literal["pickUpTip"]
DropTipCommandId = Literal["dropTip"]
MoveToSlotCommandId = Literal["moveToSlot"]
MoveToWellCommandId = Literal["moveToWell"]
DelayCommandId = Literal["delay"]
MagneticModuleEngageCommandId = Literal["magneticModule/engageMagnet"]
MagneticModuleDisengageCommandId = Literal["magneticModule/disengageMagnet"]
TemperatureModuleSetTargetCommandId = Literal["temperatureModule/setTargetTemperature"]
TemperatureModuleAwaitCommandId = Literal["temperatureModule/awaitTemperature"]
TemperatureModuleDeactivateCommandId = Literal["temperatureModule/deactivate"]
ThermocyclerSetTargetBlockCommandId = Literal["thermocycler/setTargetBlockTemperature"]
ThermocyclerSetTargetLidCommandId = Literal["thermocycler/setTargetLidTemperature"]
ThermocyclerAwaitLidTemperatureCommandId = Literal["thermocycler/awaitLidTemperature"]
ThermocyclerAwaitBlockTemperatureCommandId = Literal[
    "thermocycler/awaitBlockTemperature"
]
ThermocyclerDeactivateBlockCommandId = Literal["thermocycler/deactivateBlock"]
ThermocyclerDeactivateLidCommandId = Literal["thermocycler/deactivateLid"]
ThermocyclerOpenLidCommandId = Literal["thermocycler/openLid"]
ThermocyclerCloseLidCommandId = Literal["thermocycler/closeLid"]
ThermocyclerRunProfileCommandId = Literal["thermocycler/runProfile"]
ThermocyclerAwaitProfileCommandId = Literal["thermocycler/awaitProfileComplete"]


class NamedOffset(TypedDict):
    x: float
    y: float
    z: float


class ModuleRequirement(TypedDict):
    slot: str
    model: "ModuleModel"


class RobotRequirement(TypedDict):
    model: Literal["OT-2 Standard"]


class PipetteRequirement(TypedDict):
    mount: Union[Literal["left"], Literal["right"]]
    name: PipetteName


class PipetteAccessParams(TypedDict):
    pipette: str
    labware: str
    well: str


class FlowRateParams(TypedDict):
    flowRate: float


class PipetteAccessWithOffsetParams(PipetteAccessParams):
    offsetFromBottomMm: float


class StandardLiquidHandlingParams(PipetteAccessWithOffsetParams, FlowRateParams):
    volume: float


class AspirateCommand(TypedDict):
    command: AspirateCommandId
    params: StandardLiquidHandlingParams


class DispenseCommand(TypedDict):
    command: DispenseCommandId
    params: StandardLiquidHandlingParams


class AirGapCommand(TypedDict):
    command: AirGapCommandId
    params: StandardLiquidHandlingParams


class BlowoutParams(PipetteAccessWithOffsetParams, FlowRateParams):
    pass


class BlowoutCommand(TypedDict):
    command: BlowoutCommandId
    params: BlowoutParams


class TouchTipParams(PipetteAccessParams):
    offsetFromBottomMm: float


class TouchTipCommand(TypedDict):
    command: TouchTipCommandId
    params: TouchTipParams


class PickUpTipCommand(TypedDict):
    command: PickUpTipCommandId
    params: PipetteAccessParams


class DropTipCommand(TypedDict):
    command: DropTipCommandId
    params: PipetteAccessParams


class MoveToSlotParams(TypedDict, total=False):
    pipette: str
    slot: str
    offset: NamedOffset
    minimumZHeight: float
    forceDirect: bool


class MoveToSlotCommand(TypedDict):
    command: MoveToSlotCommandId
    params: MoveToSlotParams


class MoveToWellParams(PipetteAccessParams, total=False):
    offset: NamedOffset
    minimumZHeight: float
    forceDirect: bool


class MoveToWellCommand(TypedDict):
    command: MoveToWellCommandId
    params: MoveToWellParams


class DelayParams(TypedDict, total=False):
    wait: Union[float, bool]
    message: str


class DelayCommand(TypedDict):
    command: DelayCommandId
    params: DelayParams


class ModuleIDParams(TypedDict):
    module: str


class MagneticModuleEngageParams(ModuleIDParams):
    engageHeight: float


class MagneticModuleEngageCommand(TypedDict):
    command: MagneticModuleEngageCommandId
    params: MagneticModuleEngageParams


class MagneticModuleDisengageCommand(TypedDict):
    command: MagneticModuleDisengageCommandId
    params: ModuleIDParams


class TemperatureParams(ModuleIDParams):
    temperature: float


class TemperatureModuleSetTargetCommand(TypedDict):
    command: TemperatureModuleSetTargetCommandId
    params: TemperatureParams


class TemperatureModuleAwaitCommand(TypedDict):
    command: TemperatureModuleAwaitCommandId
    params: TemperatureParams


class TemperatureModuleDeactivateCommand(TypedDict):
    command: TemperatureModuleDeactivateCommandId
    params: ModuleIDParams


class ThermocyclerSetTargetBlockParams(TemperatureParams, total=False):
    volume: float


class ThermocyclerSetTargetBlockCommand(TypedDict):
    command: ThermocyclerSetTargetBlockCommandId
    params: ThermocyclerSetTargetBlockParams


class ThermocyclerSetTargetLidCommand(TypedDict):
    command: ThermocyclerSetTargetLidCommandId
    params: TemperatureParams


class ThermocyclerAwaitLidTemperatureCommand(TypedDict):
    command: ThermocyclerAwaitLidTemperatureCommandId
    params: TemperatureParams


class ThermocyclerAwaitBlockTemperatureCommand(TypedDict):
    command: ThermocyclerAwaitBlockTemperatureCommandId
    params: TemperatureParams


class ThermocyclerDeactivateBlockCommand(TypedDict):
    command: ThermocyclerDeactivateBlockCommandId
    params: ModuleIDParams


class ThermocyclerDeactivateLidCommand(TypedDict):
    command: ThermocyclerDeactivateLidCommandId
    params: ModuleIDParams


class ThermocyclerOpenLidCommand(TypedDict):
    command: ThermocyclerOpenLidCommandId
    params: ModuleIDParams


class ThermocyclerCloseLidCommand(TypedDict):
    command: ThermocyclerCloseLidCommandId
    params: ModuleIDParams


class ThermocyclerCycle(TypedDict):
    temperature: float
    holdTime: float


class ThermocyclerRunProfileParams(TypedDict):
    module: str
    volume: float
    profile: List[ThermocyclerCycle]


class ThermocyclerRunProfileCommand(TypedDict):
    command: ThermocyclerRunProfileCommandId
    params: ThermocyclerRunProfileParams


class ThermocyclerAwaitProfileCommand(TypedDict):
    command: ThermocyclerAwaitProfileCommandId
    params: ModuleIDParams


ThermocyclerCommand = Union[
    ThermocyclerAwaitProfileCommand,
    ThermocyclerRunProfileCommand,
    ThermocyclerCloseLidCommand,
    ThermocyclerOpenLidCommand,
    ThermocyclerDeactivateLidCommand,
    ThermocyclerDeactivateBlockCommand,
    ThermocyclerAwaitBlockTemperatureCommand,
    ThermocyclerAwaitLidTemperatureCommand,
    ThermocyclerSetTargetLidCommand,
    ThermocyclerSetTargetBlockCommand,
]
ThermocyclerCommandId = Union[
    ThermocyclerAwaitProfileCommandId,
    ThermocyclerRunProfileCommandId,
    ThermocyclerCloseLidCommandId,
    ThermocyclerOpenLidCommandId,
    ThermocyclerDeactivateLidCommandId,
    ThermocyclerDeactivateBlockCommandId,
    ThermocyclerAwaitBlockTemperatureCommandId,
    ThermocyclerAwaitLidTemperatureCommandId,
    ThermocyclerSetTargetLidCommandId,
    ThermocyclerSetTargetBlockCommandId,
]

TemperatureModuleCommand = Union[
    TemperatureModuleAwaitCommand,
    TemperatureModuleSetTargetCommand,
    TemperatureModuleDeactivateCommand,
]
TemperatureModuleCommandId = Union[
    TemperatureModuleAwaitCommandId,
    TemperatureModuleSetTargetCommandId,
    TemperatureModuleDeactivateCommandId,
]

MagneticModuleCommand = Union[
    MagneticModuleEngageCommand, MagneticModuleDisengageCommand
]
MagneticModuleCommandId = Union[
    MagneticModuleEngageCommandId, MagneticModuleDisengageCommandId
]


ModuleCommand = Union[
    ThermocyclerCommand, TemperatureModuleCommand, MagneticModuleCommand
]
ModuleCommandId = Union[
    ThermocyclerCommandId, TemperatureModuleCommandId, MagneticModuleCommandId
]

PipetteCommand = Union[
    AspirateCommand,
    DispenseCommand,
    AirGapCommand,
    BlowoutCommand,
    TouchTipCommand,
    PickUpTipCommand,
    DropTipCommand,
    MoveToSlotCommand,
    MoveToWellCommand,
]
PipetteCommandId = Union[
    AspirateCommandId,
    DispenseCommandId,
    AirGapCommandId,
    BlowoutCommandId,
    TouchTipCommandId,
    PickUpTipCommandId,
    DropTipCommandId,
    MoveToSlotCommandId,
    MoveToWellCommandId,
]

RobotCommand = Union[DelayCommand]
RobotCommandId = Union[DelayCommandId]

Command = Union[ModuleCommand, PipetteCommand, RobotCommand]
CommandId = Union[ModuleCommandId, PipetteCommandId, RobotCommandId]

V3Command = Union[PipetteCommand, RobotCommand]
V3CommandId = Union[PipetteCommandId, RobotCommandId]


class DesignerApplication(TypedDict):
    name: str
    version: str
    data: Dict[str, Any]


JsonProtocolV4 = TypedDict(
    "JsonProtocolV4",
    {
        "$otSharedSchema": Literal["#/protocol/schemas/4"],
        "schemaVersion": Literal[4],
        "metadata": Metadata,
        "robot": RobotRequirement,
        "pipettes": Dict[str, PipetteRequirement],
        "labware": Dict[str, LabwareRequirement],
        "labwareDefinitions": Dict[str, LabwareDefinition2],
        "modules": Dict[str, ModuleRequirement],
        "commands": List[Command],
        "commandAnnotations": Dict[str, Any],
        "designerApplication": DesignerApplication,
    },
    total=False,
)

JsonProtocolV5 = TypedDict(
    "JsonProtocolV5",
    {
        "$otSharedSchema": Literal["#/protocol/schemas/5"],
        "schemaVersion": Literal[5],
        "metadata": Metadata,
        "robot": RobotRequirement,
        "pipettes": Dict[str, PipetteRequirement],
        "labware": Dict[str, LabwareRequirement],
        "labwareDefinitions": Dict[str, LabwareDefinition2],
        "modules": Dict[str, ModuleRequirement],
        "commands": List[Command],
        "commandAnnotations": Dict[str, Any],
        "designerApplication": DesignerApplication,
    },
    total=False,
)


class JsonProtocolV3(TypedDict, total=False):
    schemaVersion: Literal[3]
    metadata: Metadata
    robot: RobotRequirement
    pipettes: Dict[str, PipetteRequirement]
    labware: Dict[str, LabwareRequirement]
    labwareDefinitions: Dict[str, LabwareDefinition2]
    commands: List[V3Command]
    commandAnnotations: Dict[str, Any]
    designerApplication: DesignerApplication


# todo(mm, 2022-03-18): Should this include JsonProtocolV5?
JsonProtocol = Union[JsonProtocolV4, JsonProtocolV3]



"""Fixtures and helpers for python tests."""

import pytest
from typing import Any, Dict


@pytest.fixture
def sample_transfer_properties_dict() -> Dict[str, Dict[str, Any]]:
    """A dictionary representation of transfer properties of a liquid class."""
    return {
        "flex_1channel_50": {
            "opentrons/opentrons_flex_96_tiprack_50ul/1": {
                "aspirate": {
                    "aspirate_position": {
                        "offset": {"x": 1, "y": 2, "z": 3},
                        "position_reference": "well-bottom",
                    },
                    "correction_by_volume": [(0.0, 0.0)],
                    "delay": {"enable": False},
                    "flow_rate_by_volume": [(10.0, 40.0), (20.0, 30.0)],
                    "mix": {"enable": False},
                    "pre_wet": True,
                    "retract": {
                        "air_gap_by_volume": [(5.0, 3.0), (10.0, 4.0)],
                        "delay": {"enable": False},
                        "end_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                        "speed": 40,
                        "touch_tip": {"enable": False},
                    },
                    "submerge": {
                        "delay": {"enable": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                    },
                },
                "dispense": {
                    "dispense_position": {
                        "offset": {"x": 1, "y": 2, "z": 3},
                        "position_reference": "well-bottom",
                    },
                    "correction_by_volume": [(0.0, 0.0)],
                    "delay": {"enable": False},
                    "flow_rate_by_volume": [(10.0, 40.0), (20.0, 30.0)],
                    "mix": {"enable": False},
                    "push_out_by_volume": [(10.0, 7.0), (20.0, 10.0)],
                    "retract": {
                        "air_gap_by_volume": [(5.0, 3.0), (10.0, 4.0)],
                        "blowout": {"enable": False},
                        "delay": {"enable": False},
                        "end_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                        "speed": 40,
                        "touch_tip": {"enable": False},
                    },
                    "submerge": {
                        "delay": {"enable": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 1, "y": 2, "z": 3},
                            "position_reference": "well-bottom",
                        },
                    },
                },
                "multi_dispense": {
                    "dispense_position": {
                        "offset": {"x": 0, "y": 0, "z": 1},
                        "position_reference": "well-bottom",
                    },
                    "flow_rate_by_volume": [(0, 318)],
                    "correction_by_volume": [(0, 0)],
                    "delay": {"enabled": False},
                    "submerge": {
                        "delay": {"enabled": False},
                        "speed": 100,
                        "start_position": {
                            "offset": {"x": 0, "y": 0, "z": 2},
                            "position_reference": "well-top",
                        },
                    },
                    "retract": {
                        "air_gap_by_volume": [(0, 0)],
                        "delay": {"enabled": False},
                        "end_position": {
                            "offset": {"x": 0, "y": 0, "z": 2},
                            "position_reference": "well-top",
                        },
                        "speed": 50,
                        "touch_tip": {"enabled": False},
                        "blowout": {
                            "enabled": True,
                            "location": "trash",
                            "flow_rate": 478,
                        },
                    },
                    "conditioning_by_volume": [(0, 0)],
                    "disposal_by_volume": [(0, 5)],
                },
            }
        }
    }



from enum import Enum
from typing import Any, List, Optional, Dict
from typing_extensions import Literal

from pydantic import BaseModel, Field, ConfigDict

from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.command import known_schema_ids

from .shared_models import (
    Liquid,
    Metadata,
    DesignerApplication,
    Robot,
)


class Command(BaseModel):
    commandType: str
    params: Dict[str, Any]
    key: Optional[str] = None


class CommandAnnotation(BaseModel):
    commandKeys: List[str]
    annotationType: str
    model_config = ConfigDict(extra="allow")


CommandSchemaId = Enum(  # type: ignore[misc]
    "CommandSchemaId",
    ((schema_id, schema_id) for schema_id in known_schema_ids()),
    type=str,
)


class ProtocolSchemaV8(BaseModel):
    otSharedSchema: Literal["#/protocol/schemas/8"] = Field(
        ...,
        alias="$otSharedSchema",
        description="The path to a valid Opentrons shared schema relative to "
        "the shared-data directory, without its extension.",
    )
    schemaVersion: Literal[8]
    metadata: Metadata
    robot: Robot
    liquidSchemaId: Literal["opentronsLiquidSchemaV1"]
    liquids: Dict[str, Liquid]
    labwareDefinitionSchemaId: Literal["opentronsLabwareSchemaV2"]
    labwareDefinitions: Dict[str, LabwareDefinition]
    commandSchemaId: CommandSchemaId
    commands: List[Command]
    commandAnnotationSchemaId: Literal["opentronsCommandAnnotationSchemaV1"]
    commandAnnotations: List[CommandAnnotation]
    designerApplication: Optional[DesignerApplication] = None
    model_config = ConfigDict(populate_by_name=True)
